"""Test for the `CalibrationNode` class"""

import os
from datetime import datetime
from io import StringIO
from unittest.mock import MagicMock, call, patch

import numpy as np
import pytest

from qililab.calibration.calibration_node import CalibrationNode, _json_serialize, export_nb_outputs

#################################################################################
# SET UPS ####################################
#################################################################################

logger_output_start = "RAND_INT:47102512880765720413 - OUTPUTS: "
dummy_comparison_model = MagicMock()  # Dummy comparison model, to provide to the fixtures.


####################
# MOCKED NODES ###
####################
@pytest.fixture(name="initialize_node_no_optional")
def fixture_initialize_node_no_optional() -> CalibrationNode:
    """Return a mocked CalibrationNode object for initialization, with the minimum number of things specified or mocked."""
    return CalibrationNode(
        nb_path="tests/calibration/notebook_test/zeroth.ipynb",
        qubit_index=0,
    )


@pytest.fixture(name="initialize_node_optional")
@patch(
    "qililab.calibration.calibration_node.CalibrationNode.get_last_calibrated_timestamp",
    return_value=0.0,
)
def fixture_initialize_node_optional(_) -> CalibrationNode:
    """Return a mocked CalibrationNode object for initialization, with everything specified or mocked."""
    return CalibrationNode(
        nb_path="tests/calibration/notebook_test/zeroth.ipynb",
        qubit_index=[0, 1],
        node_distinguisher=1,
        input_parameters={"a": 0, "b": 1},
        sweep_interval=np.array([0, 1, 2]),
    )


@pytest.fixture(name="methods_node")
@patch("qililab.calibration.calibration_node.CalibrationNode.get_last_calibrated_timestamp", return_value=1111)
@patch("qililab.calibration.calibration_node.StringIO", autospec=True)
def fixture_methods_node(_, __) -> CalibrationNode:
    """Return a mocked CalibrationNode object."""
    return CalibrationNode(
        nb_path="./foobar.ipynb",
        qubit_index=0,
    )


@pytest.fixture(name="methods_node_run", params=[0, [0, 2]])
@patch("qililab.calibration.calibration_node.CalibrationNode.get_last_calibrated_timestamp", return_value=1111)
@patch("qililab.calibration.calibration_node.StringIO", autospec=True)
def fixture_methods_node_run(_, __, request: pytest.FixtureRequest) -> CalibrationNode:
    """Return a mocked CalibrationNode object."""
    return CalibrationNode(
        nb_path="./foobar.ipynb",
        qubit_index=request.param,
    )


def side_efect_execute_create_file_and_raise(*args, **kwargs):
    """Generates a dummy file and raises an exception"""
    os.close(os.open("foobar_q0_dirty.ipynb", os.O_CREAT))
    raise ValueError("Test error")


#################################################################################
# TESTS FOR THE CLASS ##############################
#################################################################################


###########################
# TEST INITIALIZATION ###
###########################
class TestInitializationCalibrationNode:
    """Unit tests for the CalibrationNode class initialization."""

    def test_good_init_method_without_optional(self, initialize_node_no_optional):
        """Test a valid initialization of the class, without passing optional arguments."""
        # sourcery skip: class-extract-method
        # Assert:
        assert initialize_node_no_optional.nb_path == os.path.abspath("tests/calibration/notebook_test/zeroth.ipynb")
        assert initialize_node_no_optional.qubit_index == 0
        assert initialize_node_no_optional.node_distinguisher is None
        assert initialize_node_no_optional.node_id == "zeroth_q0"
        assert initialize_node_no_optional.nb_folder == os.path.abspath("tests/calibration/notebook_test")
        assert initialize_node_no_optional.input_parameters is None
        assert initialize_node_no_optional.sweep_interval is None
        assert initialize_node_no_optional.output_parameters is None
        assert initialize_node_no_optional.previous_timestamp is None
        assert isinstance(initialize_node_no_optional._stream, StringIO)
        assert initialize_node_no_optional.been_calibrated is False

    def test_good_init_method_with_optional(self, initialize_node_optional):
        """Test a valid initialization of the class, passing all optional arguments."""
        # Assert:
        assert initialize_node_optional.nb_path == os.path.abspath("tests/calibration/notebook_test/zeroth.ipynb")
        assert initialize_node_optional.qubit_index == [0, 1]
        assert initialize_node_optional.node_distinguisher == 1
        assert initialize_node_optional.node_id == "zeroth_1_q0q1"
        assert initialize_node_optional.nb_folder == os.path.abspath("tests/calibration/notebook_test")
        assert initialize_node_optional.input_parameters == {"a": 0, "b": 1}
        assert initialize_node_optional.sweep_interval.all() == np.array([0, 1, 2]).all()
        assert initialize_node_optional.output_parameters is None
        assert initialize_node_optional.previous_timestamp == 0.0
        assert isinstance(initialize_node_optional._stream, StringIO)
        assert initialize_node_optional.been_calibrated is False

    def test_bad_nb_path_initialization(self):
        """Test an invalid initialization of the class due to the nb_path.

        This happens when the path doesn't follow Unix format.
        """
        # Assert:
        with pytest.raises(ValueError) as error:
            _ = CalibrationNode(
                nb_path=".bizzfuzz\\foobar.ipynb",
                qubit_index=0,
            )
        assert str(error.value) == "`nb_path` must be written in unix format: `folder/subfolder/.../file.ipynb`."

    def test_bad_qubit_index_initialization(self):
        """Test an invalid initialization of the class due to `qubit_index`.

        This happens when the `qubit_index` is a list not having exactly two elements.
        """
        # Assert:
        with pytest.raises(ValueError) as error:
            _ = CalibrationNode(
                nb_path=".bizzfuzz/foobar.ipynb",
                qubit_index=[0, 1, 2],
            )
        assert str(error.value) == "List of `qubit_index` only accepts two qubit index"


class TestPublicMethodsFromCalibrationNode:
    """Unit tests for the CalibrationNode class public methods."""

    #########################
    # TEST RUN NOTEBOOK ###
    #########################

    @pytest.mark.parametrize(
        "sweep_interval, input_parameters, json_serialize_return",
        [
            (None, {"a": 0, "b": 1}, {"a": 0, "b": 1}),
            (None, None, {}),
            ([1], None, {"sweep_interval": [1]}),
            (
                [1],
                {"start": 0, "stop": 10, "step": 1},
                {"sweep_interval": [1], "start": 0, "stop": 10, "step": 1},
            ),
        ],
    )
    @patch(
        "qililab.calibration.calibration_node.CalibrationNode._execute_notebook",
        return_value={
            "platform_parameters": {"x": 0, "y": 1},
            "fidelities": {"x": 0, "y": 1},
        },
    )
    @patch(
        "qililab.calibration.calibration_node.CalibrationNode._create_notebook_datetime_path",
        return_value="",
    )
    @patch("qililab.calibration.calibration_node.os.rename")
    @patch("qililab.calibration.calibration_node.logger.error")
    @patch("qililab.calibration.calibration_node._json_serialize")
    def test_run_node(
        self,
        mock_json_serialize,
        mock_logger,
        mock_os,
        mock_create,
        mock_execute,
        sweep_interval,
        input_parameters,
        json_serialize_return,
        methods_node_run: CalibrationNode,
    ):
        """Test that run_node works properly."""
        if isinstance(methods_node_run.qubit_index, int):
            test_qubit_dict = {"qubit": methods_node_run.qubit_index}
        elif isinstance(methods_node_run.qubit_index, list):
            test_qubit_dict = {
                "control_qubit": methods_node_run.qubit_index[0],
                "target_qubit": methods_node_run.qubit_index[1],
            }
        test_params_dict = json_serialize_return | test_qubit_dict

        mock_json_serialize.return_value = test_params_dict
        methods_node_run.sweep_interval = sweep_interval
        methods_node_run.input_parameters = input_parameters
        timestamp = methods_node_run.run_node()

        mock_create.assert_has_calls([call(dirty=True), call(timestamp=timestamp)])
        mock_json_serialize.assert_called_once_with(test_params_dict)
        mock_execute.assert_called_with(methods_node_run.nb_path, "", test_params_dict)
        mock_os.assert_called_once_with("", "")
        mock_logger.assert_not_called()

        assert methods_node_run.output_parameters == mock_execute.return_value

    @pytest.mark.parametrize(
        "sweep_interval, input_parameters, json_serialize_return",
        [
            (None, None, {"qubit": 0}),
            (None, {"start": 0, "stop": 10, "step": 1}, {"qubit": 0, "start": 0, "stop": 10, "step": 1}),
            (np.array([1]), None, {"qubit": 0, "sweep_interval": [1]}),
            (
                np.array([1]),
                {"start": 0, "stop": 10, "step": 1},
                {"qubit": 0, "sweep_interval": [1], "start": 0, "stop": 10, "step": 1},
            ),
        ],
    )
    @patch("qililab.calibration.calibration_node.CalibrationNode._execute_notebook")
    @patch(
        "qililab.calibration.calibration_node.CalibrationNode._create_notebook_datetime_path",
        return_value="",
    )
    @patch("qililab.calibration.calibration_node.os.rename")
    @patch("qililab.calibration.calibration_node.logger.error")
    @patch("qililab.calibration.calibration_node._json_serialize")
    def test_run_node_interrupt(
        self,
        mock_json_serialize,
        mock_logger,
        mock_os,
        mock_create,
        mock_execute,
        sweep_interval,
        input_parameters,
        json_serialize_return,
        methods_node: CalibrationNode,
    ):
        """Test that run_node works properly when a keyboard interrupt is raised."""
        mock_json_serialize.return_value = json_serialize_return
        mock_execute.side_effect = KeyboardInterrupt()
        with pytest.raises(
            KeyboardInterrupt,
            match=f"Interrupted automatic calibration notebook execution of {methods_node.nb_path}",
        ):
            methods_node.sweep_interval = sweep_interval
            methods_node.input_parameters = input_parameters
            methods_node.run_node()

        params_dict = {"qubit": methods_node.qubit_index}

        if sweep_interval is not None:
            params_dict |= {"sweep_interval": np.array([1])}

        if input_parameters is not None:
            params_dict |= input_parameters

        mock_create.assert_called_once_with(dirty=True)
        mock_json_serialize.assert_called_once_with(params_dict)
        mock_execute.assert_called_with(methods_node.nb_path, "", params_dict)
        mock_os.assert_not_called()

        mock_logger.called_with("Interrupted automatic calibration notebook execution of %s", methods_node.nb_path)

    @pytest.mark.parametrize(
        "sweep_interval, input_parameters",
        [
            (None, {"start": 0, "stop": 10, "step": 1}),
            (None, None),
            (np.array([1]), {"start": 0, "stop": 10, "step": 1}),
            (np.array([1]), None),
        ],
    )
    @patch("qililab.calibration.calibration_node.CalibrationNode._execute_notebook")
    @patch("qililab.calibration.calibration_node.CalibrationNode._create_notebook_datetime_path", return_value="")
    @patch("qililab.calibration.calibration_node.logger.error")
    def test_run_node_raises(
        self,
        mock_logger,
        mock_create,
        mock_execute,
        sweep_interval,
        input_parameters,
        methods_node: CalibrationNode,
    ):
        """Test that run_node works properly when an exception is raised."""
        mock_execute.side_effect = ValueError("Test error")
        with pytest.raises(Exception):
            methods_node.sweep_interval = sweep_interval
            methods_node.input_parameters = input_parameters
            methods_node.run_node()

        params_dict = {"qubit": methods_node.qubit_index}

        if sweep_interval is not None:
            params_dict |= {"sweep_interval": np.array([1])}

        if input_parameters is not None:
            params_dict |= input_parameters

        mock_execute.assert_called_with(methods_node.nb_path, "", params_dict)
        assert mock_create.call_count == 1

        mock_logger.called_with(
            "Aborting execution. Exception %s during automatic calibration notebook execution, trace of the error can be found in %s",
            "Test error",
            "error_path/foobar_error.ipynb",
        )

    @pytest.mark.parametrize(
        "sweep_interval, input_parameters",
        [
            (None, {"start": 0, "stop": 10, "step": 1}),
            (None, None),
            (np.array([1]), {"start": 0, "stop": 10, "step": 1}),
            (np.array([1]), None),
        ],
    )
    @patch(
        "qililab.calibration.calibration_node.CalibrationNode._execute_notebook",
        side_effect=side_efect_execute_create_file_and_raise,
    )
    @patch(
        "qililab.calibration.calibration_node.CalibrationNode._create_notebook_datetime_path",
        return_value=f"{os.getcwd()}/foobar_q0_dirty.ipynb",
    )
    @patch("qililab.calibration.calibration_node.logger.error")
    def test_run_node_raises_error_from_execution(
        self,
        mock_logger,
        mock_create,
        mock_execute,
        sweep_interval,
        input_parameters,
        methods_node: CalibrationNode,
    ):
        """Test that run_node works properly when an exception is raised and the error notebook file is created."""
        with pytest.raises(Exception):
            methods_node.sweep_interval = sweep_interval
            methods_node.input_parameters = input_parameters
            methods_node.run_node()
        # Remove created file from the mock
        os.remove(f"{os.getcwd()}/foobar_q0_dirty.ipynb")

        params_dict = {"qubit": methods_node.qubit_index}

        if sweep_interval is not None:
            params_dict |= {"sweep_interval": np.array([1])}

        if input_parameters is not None:
            params_dict |= input_parameters

        mock_execute.assert_called_with(methods_node.nb_path, f"{os.getcwd()}/foobar_q0_dirty.ipynb", params_dict)
        assert mock_create.call_count == 2

        mock_logger.called_with(
            "Aborting execution. Exception %s during automatic calibration notebook execution, trace of the error can be found in %s",
            "Test error",
            f"{os.getcwd()}/foobar_q0_dirty.ipynb",
        )

    @patch(
        "qililab.calibration.calibration_node.CalibrationNode._execute_notebook",
        side_effect=Exception("test_exception"),
    )
    @patch("qililab.calibration.calibration_node.logger", autospec=True)
    @patch("qililab.calibration.calibration_node.os")
    def test_run_node_does_not_rename_non_existent_output(self, mocked_os, mocked_logger, _, methods_node):
        """Testing when outputs are empty received from ``execute_notebook()`."""
        with pytest.raises(
            Exception,
            match="Aborting execution. Exception test_exception during automatic calibration, expected error execution file to be created but it did not",
        ):
            methods_node.run_node()

        mocked_os.assert_not_called()
        mocked_logger.error.assert_called_with(
            "Aborting execution. Exception %s during automatic calibration, expected error execution file to be created but it did not",
            "test_exception",
        )

    ##########################################
    # TEST GET LAST CALIBRATED TIMESTAMP ###
    ##########################################
    @pytest.mark.parametrize("last_exec_output", [None, "tmp_test_foobar.ipynb"])
    @patch("qililab.calibration.calibration_node.CalibrationNode._find_last_executed_calibration")
    @patch("qililab.calibration.calibration_node.os.path.getmtime")
    def test_get_last_calibrated_timestamp(
        self, mocked_os, mock_last_exec, last_exec_output, methods_node: CalibrationNode
    ):
        """Test that ``get_last_executed_calibration()`` works correctly."""
        mock_last_exec.return_value = last_exec_output
        test_output = methods_node.get_last_calibrated_timestamp()
        mock_last_exec.assert_called_once()
        if last_exec_output is not None:
            mocked_os.assert_called_once()
        else:
            assert test_output is None


class TestPrivateMethodsFromCalibrationNode:
    """Unit tests for the CalibrationNode class private methods."""

    #############################
    # TEST EXECUTE NOTEBOOK ###
    #############################
    @pytest.mark.parametrize(
        "output",
        [
            'RAND_INT:47102512880765720413 - OUTPUTS: {"platform_parameters": [["bus_alias", "param_name", 1]]}\n',
            'RAND_INT:47102512880765720413 - OUTPUTS: {\\"platform_parameters\\": [[\\"bus_alias\\", \\"param_name\\", 1]]}\\n',
            'RAND_INT:47102512880765720413 - OUTPUTS: {\\"platform_parameters\\": [[\\"bus_alias\\", \\"param_name\\", 1]]}\\n"',
        ],
    )
    @patch("qililab.calibration.calibration_node.pm.execute_notebook")
    @patch("qililab.calibration.calibration_node.os.chdir")
    @patch("qililab.calibration.calibration_node.os.getcwd")
    def test_execute_notebook(
        self, mocked_os_getcwd, mocked_os_chdir, mocked_pm_exec, output, methods_node: CalibrationNode
    ):
        """Testing general behavior of ``execute_notebook()``."""
        # Creating expected values for assert
        expected = {"platform_parameters": [["bus_alias", "param_name", 1]]}

        # Mocking return value of stream and calling execute_notebook
        methods_node._stream.getvalue.return_value = output  # type: ignore[attr-defined]
        test_value = methods_node._execute_notebook(methods_node.nb_path, "", {})

        # Asserts
        mocked_os_getcwd.assert_called_once()
        assert mocked_os_chdir.call_count == 2
        mocked_pm_exec.assert_called_once_with(
            methods_node.nb_path, "", {}, log_output=True, stdout_file=methods_node._stream
        )
        assert test_value == expected

    @patch("qililab.calibration.calibration_node.pm.execute_notebook")
    @patch("qililab.calibration.calibration_node.os.chdir")
    @patch("qililab.calibration.calibration_node.os.getcwd")
    def test_execute_notebook_without_output(
        self, mocked_os_getcwd, mocked_os_chdir, mocked_pm_exec, methods_node: CalibrationNode
    ):
        """Testing general behavior of ``execute_notebook()``."""
        # Creating expected values for assert

        # Mocking return value of stream and calling execute_notebook
        methods_node._stream.getvalue.return_value = ""  # type: ignore[attr-defined]
        test_value = methods_node._execute_notebook(methods_node.nb_path, "", {})

        # Asserts
        mocked_os_getcwd.assert_called_once()
        assert mocked_os_chdir.call_count == 2
        mocked_pm_exec.assert_called_once_with(
            methods_node.nb_path, "", {}, log_output=True, stdout_file=methods_node._stream
        )
        assert methods_node.output_parameters == test_value is None

    @pytest.mark.parametrize(
        "output",
        [
            'RAND_INT:47102512880765720413 - OUTPUTS: {"platform_parameters": {"fizz":"buzz"}} RAND_INT:47102512880765720413 - OUTPUTS: {"platform_parameters": {"foo": "bar"}}',
            'RAND_INT:47102512880765720413 - OUTPUTS: {"platform_parameters": {"a":2}}RAND_INT:47102512880765720413 - OUTPUTS: {"platform_parameters": {"a":2}}RAND_INT:47102512880765720413 - OUTPUTS: {"platform_parameters": {"a":2}}/n',
        ],
    )
    @patch("qililab.calibration.calibration_node.pm.execute_notebook")
    @patch("qililab.calibration.calibration_node.logger", autospec=True)
    def test_execute_notebook_warnings_more_than_one_output(self, mocked_logger, mocked_pm_exec, output, methods_node):
        """Testing when no outputs or more than one outputs are received from ``execute_notebook()``."""
        methods_node._stream.getvalue.return_value = output  # type: ignore[attr-defined]

        methods_node._execute_notebook(methods_node.nb_path, "", {})

        mocked_logger.warning.assert_called_with(
            "If you had multiple outputs exported in %s, the last one found will be used.",
            methods_node.nb_path,
        )

        mocked_pm_exec.assert_called_once_with(
            methods_node.nb_path, "", {}, log_output=True, stdout_file=methods_node._stream
        )

    ##########################################
    # TEST CREATE NOTEBOOK DATETIME PATH ###
    ##########################################
    @pytest.mark.parametrize(
        "timestamp, dirty, error",
        [(None, True, True), (145783952598, False, True), (145783959532, False, False), (None, True, False)],
    )
    def test_create_notebook_datetime_path(
        self,
        timestamp,
        dirty,
        error,
        methods_node: CalibrationNode,
    ):
        """Test ``that create_notebook_datetime_path()`` works correctly."""
        with patch("qililab.calibration.calibration_node.os") as mocked_os:
            test_value = methods_node._create_notebook_datetime_path(timestamp=timestamp, dirty=dirty, error=error)
            mocked_os.makedirs.assert_called()
            if timestamp is not None:
                test_timestamp = datetime.fromtimestamp(timestamp)
                test_daily_path = f"{test_timestamp.year}_{test_timestamp.month:02d}_{test_timestamp.day:02d}"
                test_path = (
                    f"{test_daily_path}-"
                    + f"{test_timestamp.hour:02d}:{test_timestamp.minute:02d}:{test_timestamp.second:02d}"
                )
                # TODO: Solve this assert:
                # assert os.path.join(methods_node.nb_path, methods_node.node_id) in test_value
                assert f"_{test_path}" in test_value
            if dirty and not error:
                path_and_node_id = os.path.join(methods_node.nb_folder, methods_node.node_id)
                assert path_and_node_id in test_value
                assert "_dirty.ipynb" in test_value
            if error:
                path_and_node_id_error = os.path.join(methods_node.nb_folder, methods_node.node_id)
                assert path_and_node_id_error in test_value
                assert "_error.ipynb" in test_value

    ####################################
    # TEST PATH TO NAME AND FOLDER ###
    ####################################
    @pytest.mark.parametrize(
        "qubit, node_distinguisher, original_path, expected",
        [
            (0, 1, "foo/bar/foobar.ipynb", ("foobar_1_q0", "foo/bar")),
            ([0, 1], "long", "this/is/a/long/path/to/notebook.ipynb", ("notebook_long_q0q1", "this/is/a/long/path/to")),
        ],
    )
    def test_path_to_name_and_folder(
        self, methods_node: CalibrationNode, qubit, node_distinguisher, original_path, expected
    ):
        """Test that ``path_to_name_and_folder()`` works properly."""
        methods_node.qubit_index = qubit
        methods_node.node_distinguisher = node_distinguisher
        test_values = methods_node._path_to_name_and_folder(original_path)
        assert len(test_values) == 2
        assert test_values[0] == expected[0]
        assert test_values[1] == expected[1]

    ###########################################
    # TEST FIND LAST EXECUTED CALIBRATION ###
    ###########################################
    def test_find_last_executed_calibration(self, methods_node: CalibrationNode):
        """Test that ``find_last_executed_calibration()`` works correctly."""
        test_filenames = [
            "tmp_test_foobar_q0_dirty.ipynb",
            "tmp_test_foobar_q0_error.ipynb",
            "tmp_test_foo_q0_calibrated.ipynb",
            "tmp_test_bar_q0_calibrated.ipynb",
            "tmp_test_foobar_q0_.ipynb",
        ]
        filename_expected = "tmp_test_foobar_q0_calibrated.ipynb"

        for test_filename in test_filenames:
            f = open(os.path.join(methods_node.nb_folder, test_filename), "w")
            f.close()
        f = open(os.path.join(methods_node.nb_folder, filename_expected), "w")
        f.close()

        test_filename = methods_node._find_last_executed_calibration()

        assert filename_expected == test_filename

        for test_filename in test_filenames:
            os.remove(os.path.join(methods_node.nb_folder, test_filename))
        os.remove(os.path.join(methods_node.nb_folder, filename_expected))

    def test_find_last_executed_calibration_does_not_find_file(self, methods_node: CalibrationNode):
        """Test ``find_last_executed_calibration()`` works properly, when there is nothing to find."""
        test_filenames = [
            "tmp_test_foobar_dirty_q0.ipynb",
            "tmp_test_foobar_error_q0.ipynb",
            "tmp_test_foo_calibrated_q0.ipynb",
            "tmp_test_bar_calibrated_q0.ipynb",
            "tmp_test_foobar_.ipynb_q0",
        ]

        for test_filename in test_filenames:
            f = open(os.path.join(methods_node.nb_folder, test_filename), "w")
            f.close()

        test_filename = methods_node._find_last_executed_calibration()

        assert test_filename is None

        for test_filename in test_filenames:
            os.remove(os.path.join(methods_node.nb_folder, test_filename))

    ##########################################
    # TEST ADD STRING TO CHECKED NB NAME ###
    ##########################################
    def test_add_string_to_checked_nb_name(self, methods_node: CalibrationNode):
        """Test that ``add_string_to_checked_nb_name()`` works properly."""
        with patch("qililab.calibration.calibration_node.os.rename") as mocked_rename:
            timestamp_path = methods_node._create_notebook_datetime_path(timestamp=0).split(".ipynb")[0]
            string_to_add = "test_succesful"
            methods_node._add_string_to_checked_nb_name(string_to_add, timestamp=0)
            mocked_rename.assert_called_once_with(f"{timestamp_path}.ipynb", f"{timestamp_path}_{string_to_add}.ipynb")


class TestStaticMethodsFromCalibrationNode:
    """Test static methods of the `CalibrationNode` class."""

    ##########################################
    # TEST BUILD NOTEBOOKS LOGGER STREAM ###
    ##########################################
    @patch("qililab.calibration.calibration_node.logging", autospec=True)
    def test_build_notebooks_logger_stream(self, mocked_logging):
        """Test that ``build_notebooks_logger_stream()`` works properly."""
        stream = CalibrationNode._build_notebooks_logger_stream()
        mocked_logging.basicConfig.assert_called_once()
        assert isinstance(stream, StringIO)


#################################################################################
# TESTS FOR THE EXTERNAL FUNCTIONS #######################
#################################################################################


#######################################
# TEST EXPORT CALIBRATION OUTPUTS ###
#######################################
@pytest.mark.parametrize(
    "test_outputs, test_serializable_outputs",
    [
        ({"this_is": "a_test_dict", "foo": [1, 2, 3, 4]}, {"this_is": "a_test_dict", "foo": [1, 2, 3, 4]}),
        (
            {"this_is": np.array([1, 2, 3, 4, 5]), "foo": {"bar": "jose", "pepe": (np.array([0]), np.array([0]), "a")}},
            {"this_is": [1, 2, 3, 4, 5], "foo": {"bar": "jose", "pepe": [[0], [0], "a"]}},
        ),
    ],
)
@patch("qililab.calibration.calibration_node.json.dumps", autospec=True)
def test_export_nb_outputs(mocked_dumps, test_outputs, test_serializable_outputs):
    """Test that ``export_nb_outputs()`` works properly."""
    mocked_dumps.return_value = test_serializable_outputs
    with patch("builtins.print") as mocked_print:
        export_nb_outputs(test_outputs)
        mocked_dumps.assert_called_with(test_serializable_outputs)
        mocked_print.assert_called_with(f"{logger_output_start}{test_serializable_outputs}")


#######################################
# TEST JSON SERIALIZE #########
#######################################
@pytest.mark.parametrize(
    "test_objects, expected_test_objects",
    [
        (
            [np.array([1, 2, 3, 4]), [np.array([1, 2, 3, 4]), [2 + 1j, np.complex128(2 + 1j)]]],
            [[1.0, 2.0, 3.0, 4.0], [[1.0, 2.0, 3.0, 4.0], [{"real": 2.0, "imag": 1.0}, {"real": 2.0, "imag": 1.0}]]],
        ),
        (
            [np.True_, [np.float64(12.9), np.void(5)]],
            [True, [12.9, None]],
        ),
        (
            {"this_is": "a_test_dict", "foo": np.array([1, 2, 3, 4])},
            {"this_is": "a_test_dict", "foo": [1.0, 2.0, 3.0, 4.0]},
        ),
        (
            {"this_is": np.array([1, 2, 3, 4, 5]), "foo": {"bar": "jose", "pepe": (np.array([0]), np.array([0]), "a")}},
            {"this_is": [1.0, 2.0, 3.0, 4.0, 5.0], "foo": {"bar": "jose", "pepe": [[0], [0], "a"]}},
        ),
        (123, 123.0),
        (
            (np.array([1, 2, 3, 4, 5.5]), {"foo": np.array([0.1, 0.2, 0.3])}),
            [[1.0, 2.0, 3.0, 4.0, 5.5], {"foo": [0.1, 0.2, 0.3]}],
        ),
        (np.array([20, 30, 40]), [20.0, 30.0, 40.0]),
        ("qililab rocks!", "qililab rocks!"),
    ],
)
def test_json_serialize(test_objects, expected_test_objects):
    """Test that ``json_serialize()`` works properly."""
    test_result = _json_serialize(test_objects)
    assert test_result == expected_test_objects



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""KeySight Vector Network Analyzer E5080B class."""

import time
from dataclasses import dataclass

import numpy as np

from qililab.constants import DEFAULT_TIMEOUT
from qililab.instruments.decorators import log_set_parameter
from qililab.instruments.instrument import ParameterNotFound
from qililab.instruments.utils import InstrumentFactory
from qililab.instruments.vector_network_analyzer import VectorNetworkAnalyzer
from qililab.result.vna_result import VNAResult
from qililab.typings import ChannelID, InstrumentName, Parameter, ParameterValue
from qililab.typings.enums import VNASweepModes
from qililab.typings.instruments.vector_network_analyzer import VectorNetworkAnalyzerDriver


@InstrumentFactory.register
class E5080B(VectorNetworkAnalyzer):
    """KeySight Vector Network Analyzer E5080B"""

    name = InstrumentName.KEYSIGHT_E5080B
    device: VectorNetworkAnalyzerDriver

    @dataclass
    class E5080BSettings(VectorNetworkAnalyzer.VectorNetworkAnalyzerSettings):
        """Contains the settings of a specific VectorNetworkAnalyzer.

        Args:
            sweep_mode (str): Sweeping mode of the instrument
        """

        sweep_mode: VNASweepModes = VNASweepModes.CONT
        device_timeout: float = DEFAULT_TIMEOUT

    settings: E5080BSettings

    @log_set_parameter
    def set_parameter(self, parameter: Parameter, value: ParameterValue, channel_id: int = 1, port: int = 1) -> None:
        """Get instrument parameter.

        Args:
            parameter (Parameter): Name of the parameter to get.
            channel_id (int): Channel identifier of the parameter to update.
            port (int): Port identifier of the parameter to update.
        """
        if parameter == Parameter.POWER:
            value = float(value)
            self.set_power(power=value, channel=channel_id, port=port)
            return
        if parameter == Parameter.IF_BANDWIDTH:
            value = float(value)
            self.set_if_bandwidth(value=value, channel=channel_id)
            return
        if parameter == Parameter.ELECTRICAL_DELAY:
            self.electrical_delay = value
            return
        if parameter == Parameter.SWEEP_MODE:
            self.set_sweep_mode(value=value, channel=channel_id)  # type: ignore
            return
        if parameter == Parameter.DEVICE_TIMEOUT:
            self.device_timeout = value
            return
        raise ParameterNotFound(self, parameter)

    def get_parameter(self, parameter: Parameter, channel_id: ChannelID | None = None):
        """Get instrument parameter.

        Args:
            parameter (Parameter): Name of the parameter to get.
            channel_id (int | None): Channel identifier of the parameter to update.
        """
        if parameter == Parameter.POWER:
            return self.settings.power
        if parameter == Parameter.IF_BANDWIDTH:
            return self.settings.if_bandwidth
        if parameter == Parameter.ELECTRICAL_DELAY:
            return self.settings.electrical_delay
        if parameter == Parameter.SWEEP_MODE:
            if channel_id:
                return self._get_sweep_mode(channel=channel_id)
            return self._get_sweep_mode()
        if parameter == Parameter.DEVICE_TIMEOUT:
            return self.device_timeout
        raise ParameterNotFound(self, parameter)

    def _set_parameter_float(self, parameter: Parameter, value: float):
        """Set instrument settings parameter to the corresponding value

        Args:
            parameter (Parameter): settings parameter to be updated
            value (float): new value
        """
        if parameter == Parameter.DEVICE_TIMEOUT:
            self.device_timeout = value
            return

        super()._set_parameter_float(parameter, value)

    def _set_parameter_str(self, parameter: Parameter, value: str):
        """Set instrument settings parameter to the corresponding value

        Args:
            parameter (Parameter): settings parameter to be updated
            value (str): new value
        """
        if parameter == Parameter.SWEEP_MODE:
            self.set_sweep_mode(VNASweepModes(value))
            return

        super()._set_parameter_str(parameter, value)

    def set_power(self, power: float, channel=1, port=1):
        """sets the power in dBm"""
        self.settings.power = power
        if self.is_device_active():
            self.send_command(f"SOUR{channel}:POW{port}", f"{self.settings.power:.1f}")

    def set_if_bandwidth(self, value: float, channel=1):
        """sets the if bandwidth in Hz"""
        self.settings.if_bandwidth = value
        if self.is_device_active():
            bandwidth = str(self.settings.if_bandwidth)
            self.send_command(f"SENS{channel}:BWID", bandwidth)

    @VectorNetworkAnalyzer.electrical_delay.setter  # type: ignore
    def electrical_delay(self, value: float):
        """
        Set electrical delay in channel 1

        Input:
            value (str) : Electrical delay in ns
        """
        self.settings.electrical_delay = value
        if self.is_device_active():
            etime = f"{self.settings.electrical_delay:.12f}"
            self.send_command("SENS1:CORR:EXT:PORT1:TIME", etime)

    def get_sweep_mode(self):
        """VectorNetworkAnalyzer'sweep_mode' property.

        Returns:mode
            str: settings.sweep_mode.
        """
        return self.settings.sweep_mode

    def set_sweep_mode(self, value: str, channel=1):
        """
        Sets the sweep mode

        Input:
            mode (str) : Sweep mode: 'hold', 'cont', single' and 'group'
        """
        self.settings.sweep_mode = VNASweepModes(value)
        if self.is_device_active():
            mode = self.settings.sweep_mode.name
            self.send_command(f"SENS{channel}:SWE:MODE", mode)

    @property
    def device_timeout(self):
        """VectorNetworkAnalyzer 'device_timeout' property.

        Returns:
            float: settings.device_timeout.
        """
        return self.settings.device_timeout

    @device_timeout.setter
    def device_timeout(self, value: float):
        """sets the device timeout in mili seconds"""
        self.settings.device_timeout = value

    def _get_sweep_mode(self, channel=1):
        """Return the current sweep mode."""
        return str(self.send_query(f":SENS{channel}:SWE:MODE?")).rstrip()

    def _get_trace(self, channel=1, trace=1):
        """Get the data of the current trace."""
        self.send_command(command="FORM:DATA", arg="REAL,32")
        self.send_command(command="FORM:BORD", arg="SWAPPED")  # SWAPPED
        data = self.send_binary_query(f"CALC{channel}:MEAS{trace}:DATA:SDAT?")
        datareal = np.array(data[::2])  # Elements from data starting from 0 iterating by 2
        dataimag = np.array(data[1::2])  # Elements from data starting from 1 iterating by 2

        return datareal + 1j * dataimag

    def _set_count(self, count: str, channel=1):
        """
        Sets the trigger count (groups)
        Input:
            count (str) : Count number
        """
        self.send_command(f"SENS{channel}:SWE:GRO:COUN", count)

    def _pre_measurement(self):
        """
        Set everything needed for the measurement
        Averaging has to be enabled.
        Trigger count is set to number of averages
        """
        if not self.averaging_enabled:
            self.averaging_enabled = True
            self.number_averages = 1
        self._set_count(str(self.settings.number_averages))

    def _start_measurement(self, channel=1):
        """
        This function is called at the beginning of each single measurement in the spectroscopy script.
        Also, the averages need to be reset.
        """
        self.average_clear()
        self.settings.sweep_mode = VNASweepModes("group")
        mode = self.settings.sweep_mode.name
        self.send_command(f"SENS{channel}:SWE:MODE", mode)

    def _wait_until_ready(self, period=0.25) -> bool:
        """Waiting function to wait until VNA is ready."""
        timelimit = time.time() + self.device_timeout
        while time.time() < timelimit:
            if self.ready():
                return True
            time.sleep(period)
        return False

    def average_clear(self, channel=1):
        """Clears the average buffer."""
        self.send_command(command=f":SENS{channel}:AVER:CLE", arg="")

    def get_frequencies(self):
        """return freqpoints"""
        return np.array(self.send_binary_query("SENS:X?"))

    def ready(self) -> bool:
        """
        This is a proxy function.
        Returns True if the VNA is on HOLD after finishing the required number of averages.
        """
        try:  # the VNA sometimes throws an error here, we just ignore it
            return self._get_sweep_mode() == "HOLD"
        except Exception:  # noqa: BLE001
            return False

    def release(self, channel=1):
        """Bring the VNA back to a mode where it can be easily used by the operator."""
        self.settings.sweep_mode = VNASweepModes("cont")
        mode = self.settings.sweep_mode.name
        self.send_command(f"SENS{channel}:SWE:MODE", mode)

    def read_tracedata(self):
        """
        Return the current trace data.
        It already releases the VNA after finishing the required number of averages.
        """
        self._pre_measurement()
        self._start_measurement()
        if self._wait_until_ready():
            trace = self._get_trace()
            self.release()
            return trace
        raise TimeoutError("Timeout waiting for trace data")

    def acquire_result(self):
        """Convert the data received from the device to a Result object."""
        return VNAResult(data=self.read_tracedata())



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Platform class."""

from __future__ import annotations

import ast
import io
import re
import tempfile
from contextlib import contextmanager
from copy import deepcopy
from dataclasses import asdict
from typing import TYPE_CHECKING, Callable, cast

import numpy as np
from qibo.gates import M
from qibo.models import Circuit
from qm import generate_qua_script
from ruamel.yaml import YAML

from qililab.analog import AnnealingProgram
from qililab.config import logger
from qililab.constants import FLUX_CONTROL_REGEX, GATE_ALIAS_REGEX, RUNCARD
from qililab.digital import CircuitTranspiler
from qililab.exceptions import ExceptionGroup
from qililab.instrument_controllers import InstrumentController, InstrumentControllers
from qililab.instrument_controllers.utils import InstrumentControllerFactory
from qililab.instruments.instrument import Instrument
from qililab.instruments.instruments import Instruments
from qililab.instruments.qblox import QbloxModule
from qililab.instruments.qblox.qblox_draw import QbloxDraw
from qililab.instruments.quantum_machines import QuantumMachinesCluster
from qililab.instruments.utils import InstrumentFactory
from qililab.platform.components.bus import Bus
from qililab.platform.components.buses import Buses
from qililab.pulse.pulse_schedule import PulseSchedule
from qililab.pulse.qblox_compiler import ModuleSequencer
from qililab.pulse.qblox_compiler import QbloxCompiler as PulseQbloxCompiler
from qililab.qprogram import (
    Calibration,
    Domain,
    Experiment,
    QbloxCompilationOutput,
    QbloxCompiler,
    QProgram,
    QuantumMachinesCompilationOutput,
    QuantumMachinesCompiler,
)
from qililab.qprogram.crosstalk_matrix import CrosstalkMatrix, FluxVector
from qililab.qprogram.experiment_executor import ExperimentExecutor
from qililab.result.qblox_results.qblox_result import QbloxResult
from qililab.result.qprogram.qprogram_results import QProgramResults
from qililab.result.qprogram.quantum_machines_measurement_result import QuantumMachinesMeasurementResult
from qililab.typings import ChannelID, InstrumentName, Parameter, ParameterValue
from qililab.utils import hash_qpy_sequence

if TYPE_CHECKING:
    from queue import Queue

    from qpysequence import Sequence as QpySequence

    from qililab.digital import DigitalTranspilationConfig
    from qililab.instrument_controllers.instrument_controller import InstrumentController
    from qililab.instruments.instrument import Instrument
    from qililab.result import Result
    from qililab.settings import Runcard
    from qililab.settings.digital.gate_event_settings import GateEventSettings


class Platform:
    """Platform object representing the laboratory setup used to control quantum devices.

    The platform is responsible for managing the initializations, connections, setups, and executions of the laboratory, which mainly consists of:

    - :class:`.Chip`

    - Buses

    - Instruments

    .. note::

        This class should be instantiated with the :meth:`ql.build_platform()` function, either by passing a :ref:`runcard <runcards>` (serialized platform dictionary)
        or a path to the location of the YAML file containing it.

        More information about the runcard structure, in the :ref:`Runcards <runcards>` section of the documentation.

    After initializing a :class:`Platform`, the typical first three steps (which are usually only required at the start) are:

    >>> platform.connect()  # Connects to all the instruments.
    >>> platform.initial_setup()  # Sets the parameters defined in the runcard.
    >>> platform.turn_on_instruments()  # Turns on the signal outputs.

    And then, for each experiment you want to run, you would typically repeat:

    >>> platform.set_parameter(...)  # Sets any parameter of the Platform.
    >>> result = platform.execute(...)  # Executes the platform.

    Args:
        runcard (Runcard): Dataclass containing the serialized platform (chip, instruments, buses...), created during :meth:`ql.build_platform()` with the given runcard dictionary.

    Examples:

        .. note::

            The following examples contain made up results. These will soon be updated with real results.

        .. note::

            All the following examples are explained in detail in the :ref:`Platform <platform>` section of the documentation. However, here are a few thing to keep in mind:

            - To connect, your computer must be in the same network of the instruments specified in the runcard, with their IP's addresses. Connection is necessary for the subsequent steps.

            - You might want to skip the ``platform.initial_setup()`` and the ``platform.turn_on_instruments()`` steps if you think nothing has been modified, but we recommend doing them every time.

            - ``platform.turn_on_instruments()`` is used to turn on the signal output of all the sources defined in the runcard (RF, Voltage and Current sources).

            - You can print ``platform.chip`` and ``platform.buses`` at any time to check the platform's structure.

        **1. Executing a circuit with Platform:**


        To execute a circuit you first need to define your circuit, for example, one with a pi pulse and a measurement gate in qubit ``q`` (``int``).
        Then you also need to build, connect, set up, and execute the platform, which together look like:

        .. code-block:: python

            import qililab as ql

            from qibo.models import Circuit
            from qibo import gates

            # Defining the Rabi circuit:
            circuit = Circuit(q + 1)
            circuit.add(gates.X(q))
            circuit.add(gates.M(q))

            # Building the platform:
            platform = ql.build_platform(runcard="runcards/galadriel.yml")

            # Connecting and setting up the platform:
            platform.connect()
            platform.initial_setup()
            platform.turn_on_instruments()

            # Executing the platform:
            result = platform.execute(program=circuit, num_avg=1000, repetition_duration=6000)

        The results would look something like this:

        >>> result.array
        array([[6.],
                [6.]])

        .. note::

            The obtained values correspond to the integral of the I/Q signals received by the digitizer.
            And they have shape `(#sequencers, 2, #bins)`, in this case you only have 1 sequencer and 1 bin.

        You could also get the results in a more standard format, as already classified ``counts`` or ``probabilities`` dictionaries, with:

        >>> result.counts
        {'0': 501, '1': 499}

        >>> result.probabilities
        {'0': .501, '1': .499}

        .. note::

            You can find more information about the results, in the :class:`.Results` class documentation.

        |

        **2. Running a Rabi sweep with Platform:**

        To perform a Rabi sweep, you need the previous circuit, and again, you also need to build, connect and setup the platform.
        But this time, instead than executing the circuit once, you will loop changing the amplitude parameter of the AWG (generator of the pi pulse):

        .. code-block:: python

            # Looping over the AWG amplitude to execute the Rabi sweep:
            results = []
            amp_values = [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.9, 1.0]

            for amp in amp_values:
                platform.set_parameter(alias="drive_q", parameter=ql.Parameter.AMPLITUDE, value=amp)
                result = platform.execute(program=circuit, num_avg=1000, repetition_duration=6000)
                results.append(result.array)

        Now you can use ``np.hstack`` to stack the results horizontally. By doing this, you would obtain an
        array with shape `(2, N)`, where N is the number of elements inside the loop:

        >>> import numpy as np
        >>> np.hstack(results)
        array([[5, 4, 3, 2, 1, 2, 3, 4, 5, 4, 3],
                [5, 4, 3, 2, 1, 2, 3, 4, 5, 4, 3]])

        You can see how the integrated I/Q values oscillate, indicating that qubit ``q`` oscillates between the ground and
        excited states!

        |

        **3. A faster Rabi sweep, translating the circuit to pulses:**

        Since you are looping over variables that are independent of the circuit (in this case, the amplitude of the AWG),
        you can speed up the experiment by translating the circuit into pulses beforehand, only once, and then, executing the obtained
        pulses inside the loop.

        Which is the same as before, but passing the ``pulse_schedule`` instead than the ``circuit``, to the ``execute()`` method:

        .. code-block:: python

            from qililab.pulse.circuit_to_pulses import CircuitToPulses

            # Translating the circuit to pulses:
            pulse_schedule = CircuitToPulses(platform=platform).translate(circuits=[circuit])

            # Looping over the AWG amplitude to execute the Rabi sweep:
            results = []
            amp_values = [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.9, 1.0]

            for amp in amp_values:
                platform.set_parameter(alias="drive_q", parameter=ql.Parameter.AMPLITUDE, value=amp)
                result = platform.execute(program=pulse_schedule, num_avg=1000, repetition_duration=6000)
                results.append(result.array)

        If you now stack and print the results, you will obtain similar results, but much faster!

        >>> np.hstack(results)
        array([[5, 4, 3, 2, 1, 2, 3, 4, 5, 4, 3],
                [5, 4, 3, 2, 1, 2, 3, 4, 5, 4, 3]])
        TODO: !!! Change this results for the actual ones !!!

        |

        **4. Ramsey sequence, looping over a parameter inside the circuit:**

        To run a Ramsey sequence you also need to build, connect and set up the platform as before. However, the circuit will be different from the previous one,
        and also, this time, you need to loop over a parameter of the circuit itself, specifically over the time of the ``Wait`` gate.

        To do this, since the parameter is inside the Qibo circuit, you will need to use Qibo own ``circuit.set_parameters()`` method, specifying the
        parameters you want to set, in the same order they appear in the circuit construction:

        .. code-block:: python

            import qililab as ql

            from qibo.models import Circuit
            from qibo import gates

            # Building the platform:
            platform = ql.build_platform(runcard="runcards/galadriel.yml")

            # Connecting and setting up the platform:
            platform.connect()
            platform.initial_setup()
            platform.turn_on_instruments()

            # Defining the Ramsey circuit:
            circuit = Circuit(q + 1)
            circuit.add(gates.RX(q, theta=np.pi / 2))
            circuit.add(ql.Wait(q, t=0))
            circuit.add(gates.RX(q, theta=np.pi / 2))
            circuit.add(gates.M(q))

            # Looping over the wait time t to execute the Ramsey:
            results = []
            wait_times = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

            for wait in wait_times:
                circuit.set_parameters([np.pi / 2, wait, np.pi / 2])
                result = platform.execute(program=circuit, num_avg=1000, repetition_duration=6000)
                results.append(result.array)

        which for each execution, would set ``np.pi/2`` to the ``theta`` parameters of the ``RX`` gates, and the looped ``wait`` time  to the ``t`` parameter of the
        ``Wait`` gate.

        If you print the results, you'll see how you obtain the sinusoidal expected behaviour!

        >>> results = np.hstack(results)
        >>> results
        array([[5, 4, 3, 2, 1, 2, 3, 4, 5, 4, 3],
                [5, 4, 3, 2, 1, 2, 3, 4, 5, 4, 3]])
        TODO: !!! Change this results for the actual sinusoidal ones (change wait_times of execution if needed) !!!
    """

    def __init__(self, runcard: Runcard):
        self.name = runcard.name
        """Name of the platform (``str``) """

        self.instruments = Instruments(elements=self._load_instruments(instruments_dict=runcard.instruments))
        """All the instruments of the platform and their necessary settings (``dataclass``). Each individual instrument is contained in a list within the dataclass."""

        self.instrument_controllers = InstrumentControllers(
            elements=self._load_instrument_controllers(instrument_controllers_dict=runcard.instrument_controllers)
        )
        """All the instrument controllers of the platform and their necessary settings (``dataclass``). Each individual instrument controller is contained in a list within the dataclass."""

        self.buses = Buses(
            elements=[Bus(settings=asdict(bus), platform_instruments=self.instruments) for bus in runcard.buses]
        )
        """All the buses of the platform and their necessary settings (``dataclass``). Each individual bus is contained in a list within the dataclass."""

        self.digital_compilation_settings = runcard.digital
        """Gate settings and definitions (``dataclass``). These setting contain how to decompose gates into pulses."""

        self.analog_compilation_settings = runcard.analog
        """Flux to bus mapping for analog control"""

        self._connected_to_instruments: bool = False
        """Boolean indicating the connection status to the instruments. Defaults to False (not connected)."""

        self._qpy_sequence_cache: dict[str, str] = {}
        """Dictionary for caching qpysequences."""

        self.experiment_results_base_path: str = tempfile.gettempdir()
        """Base path for saving experiment results."""

        self.experiment_results_path_format: str = "{date}/{time}/{label}.h5"
        """Format of the experiment results path."""

        self.crosstalk: CrosstalkMatrix | None = None
        """Crosstalk matrix information, defaults to None (only used on FLUX parameters)"""

        self.flux_vector: FluxVector | None = None
        """Flux vector information, defaults to None (only used on FLUX parameters)"""

        self.flux_parameter: dict[str, int | float | bool | str] = {}
        """Flux dictionary with information for the get parameter (only used on FLUX parameters)"""

    def connect(self):
        """Connects to all the instruments and blocks the connection for other users.

        You must be connected in order to set up and turn on instruments, or in order to execute the platform.

        To connect, your computer must be in the same network of the instruments specified in the :ref:`runcard <runcards>` (with their corresponding IP's addresses).
        """
        if self._connected_to_instruments:
            logger.info("Already connected to the instruments")
            return

        self.instrument_controllers.connect()
        self._connected_to_instruments = True
        logger.info("Connected to the instruments")

    def initial_setup(self):
        """Sets the values of the cache of the :class:`.Platform` object to the connected instruments.

        If called after a ``ql.build_platform()``, where the :class:`.Platform` object is built with the provided runcard,
        this function sets the values of the :ref:`runcard <runcards>` into the connected instruments.

        It is recommended to use this function after a ``ql.build_platform()`` + ``platform.connect()`` to ensure that no parameter
        differs from the current runcard settings.

        If a `platform.set_parameter()` is called between platform building and initial setup, the value set in the instruments
        will be the new "set" value, as the cache values of the :class:`.Platform` object are modified.
        """
        if not self._connected_to_instruments:
            raise AttributeError("Can not do initial_setup without being connected to the instruments.")
        self.instrument_controllers.initial_setup()
        logger.info("Initial setup applied to the instruments")

    def turn_on_instruments(self):
        """Turns on the signal output for the generator instruments (RF, voltage sources and current sources).

        This does not actually turn on the laboratory instruments, it only opens the signal output generation of the sources.

        We recommend you to do this always after a connection and a setup, to ensure that everything is ready for an execution.
        """
        self.instrument_controllers.turn_on_instruments()
        logger.info("Instruments turned on")

    def turn_off_instruments(self):
        """Turns off the signal output for the generator instruments (local oscillators, voltage sources and current sources).

        This does not actually turn the laboratory instruments off, it only closes their signal output generation.
        """
        self.instrument_controllers.turn_off_instruments()
        logger.info("Instruments turned off")

    def disconnect(self):
        """Closes the connection to all the instruments."""
        if not self._connected_to_instruments:
            logger.info("Already disconnected from the instruments")
            return
        self.instrument_controllers.disconnect()
        self._connected_to_instruments = False
        self._qpy_sequence_cache = {}
        logger.info("Disconnected from instruments")

    def get_element(self, alias: str):
        """Gets the platform element and to which bus it is connected, using its alias.

        Args:
            alias (str): Element alias to identify it.

        Returns:
            tuple[object, list | None]: Element class together with the index of the bus where the element is located.
        """
        # TODO: fix docstring, bus is not returned in most cases
        regex_match = re.search(GATE_ALIAS_REGEX, alias.split("_")[0])
        if regex_match is not None:
            name = regex_match["gate"]
            qubits_str = regex_match["qubits"]
            qubits = ast.literal_eval(qubits_str)
            if (
                self.digital_compilation_settings is not None
                and f"{name}({qubits_str})" in self.digital_compilation_settings.gate_names
            ):
                return self.digital_compilation_settings.get_gate(name=name, qubits=qubits)
        regex_match = re.search(FLUX_CONTROL_REGEX, alias)
        if regex_match is not None:
            element_type = regex_match.lastgroup
            element_shorthands = {"qubit": "q", "coupler": "c"}
            flux = regex_match["flux"]
            # TODO: support commuting the name of the coupler eg. c1_0 = c0_1
            bus_alias = next(
                (
                    element.bus
                    for element in self.analog_compilation_settings.flux_control_topology  # type: ignore[union-attr]
                    if self.analog_compilation_settings
                    and element.flux == f"{flux}_{element_shorthands[element_type]}{regex_match[element_type]}"  # type: ignore[index]
                ),
                None,
            )
            if bus_alias is not None:
                return self.buses.get(alias=bus_alias)

        element = self.instruments.get_instrument(alias=alias)
        if element is None:
            element = self.instrument_controllers.get_instrument_controller(alias=alias)
        if element is None:
            element = self.buses.get(alias=alias)
        return element

    def _get_bus_by_alias(self, alias: str) -> Bus | None:
        """Gets buses given their alias.

        Args:
            alias (str, optional): Bus alias to identify it. Defaults to None.

        Returns:
            :class:`Bus`: Bus corresponding to the given alias. If none is found `None` is returned.

        """
        return self.buses.get(alias=alias)

    def get_parameter(self, alias: str, parameter: Parameter, channel_id: ChannelID | None = None):
        """Get platform parameter.

        Args:
            parameter (Parameter): Name of the parameter to get.
            alias (str): Alias of the bus where the parameter is set.
            channel_id (int, optional): ID of the channel we want to use to set the parameter. Defaults to None.
        """
        regex_match = re.search(GATE_ALIAS_REGEX, alias)
        if alias == "platform" or parameter == Parameter.DELAY or regex_match is not None:
            if self.digital_compilation_settings is None:
                raise ValueError("Trying to get parameter of gates settings, but no gates settings exist in platform.")
            return self.digital_compilation_settings.get_parameter(
                alias=alias, parameter=parameter, channel_id=channel_id
            )
        if parameter == Parameter.FLUX:
            if alias not in self.flux_parameter:
                self.flux_parameter[alias] = 0.0
            return self.flux_parameter[alias]
        element = self.get_element(alias=alias)
        return element.get_parameter(parameter=parameter, channel_id=channel_id)

    def _data_draw(self):
        """From the runcard retrieve the parameters necessary to draw the qprogram."""
        data_oscilloscope = {}
        buses = list(self.buses)
        instruments = {
            instrument for bus in buses for instrument in bus.instruments if isinstance(instrument, (QbloxModule))
        }
        if all(isinstance(instrument, QbloxModule) for instrument in instruments):
            for bus in buses:
                for instrument, _ in zip(bus.instruments, bus.channels):
                    if isinstance(instrument, QbloxModule):
                        data_oscilloscope[bus.alias] = {}
                        dac_offset_i = 0
                        dac_offset_q = 0
                        parameters = [
                            Parameter.IF,
                            Parameter.GAIN_I,
                            Parameter.GAIN_Q,
                            Parameter.OFFSET_I,
                            Parameter.OFFSET_Q,
                            Parameter.HARDWARE_MODULATION,
                        ]
                        for parameter in parameters:
                            data_oscilloscope[bus.alias][parameter.value] = self.get_parameter(bus.alias, parameter)

                        data_oscilloscope[bus.alias]["instrument_name"] = instrument.name.value

                        if instrument.name == InstrumentName.QBLOX_QCM or instrument.name == InstrumentName.QBLOX_QRM:
                            # retrieve the dac offset and assign it to the bus
                            identifier = bus.channels
                            for awg in instrument.awg_sequencers:
                                if awg.identifier == identifier[0]:
                                    for idx, out in enumerate(awg.outputs):
                                        if idx == 0:
                                            dac_offset_i = instrument.out_offsets[out]
                                        elif idx == 1:
                                            dac_offset_q = instrument.out_offsets[out]
                            data_oscilloscope[bus.alias]["dac_offset_i"] = dac_offset_i
                            data_oscilloscope[bus.alias]["dac_offset_q"] = dac_offset_q

                        elif instrument.name == InstrumentName.QCMRF or instrument.name == InstrumentName.QRMRF:
                            # retrieve the dac offset and assign it to the bus
                            identifier = bus.channels
                            for awg in instrument.awg_sequencers:
                                if awg.identifier == identifier[0]:
                                    for out in awg.outputs:
                                        if out == 0:
                                            dac_offset_i = bus.get_parameter(Parameter.OUT0_OFFSET_PATH0)
                                            dac_offset_q = bus.get_parameter(Parameter.OUT0_OFFSET_PATH1)
                                        elif out == 1:
                                            dac_offset_i = bus.get_parameter(Parameter.OUT1_OFFSET_PATH0)
                                            dac_offset_q = bus.get_parameter(Parameter.OUT1_OFFSET_PATH1)
                            data_oscilloscope[bus.alias]["dac_offset_i"] = dac_offset_i
                            data_oscilloscope[bus.alias]["dac_offset_q"] = dac_offset_q
        else:
            # TODO: the same information could be generated with a Quantum Machine runcard, even if the QBlox Compiler is used in the background.
            raise NotImplementedError("The drawing feature is currently only supported for QBlox.")

        return data_oscilloscope

    def set_parameter(
        self,
        alias: str,
        parameter: Parameter,
        value: ParameterValue,
        channel_id: ChannelID | None = None,
    ):
        """Set a parameter for a platform element.

        If connected to an instrument, this function updates both the cache of the :class:`.Platform` object and the
        instrument's value. Otherwise, it only stores the value in the cache. Subsequent ``connect()`` + ``initial_setup()``
        will apply the cached values into the real instruments.

        If you use ``set_parameter`` + ``ql.save_platform()``, the saved runcard will include the new "set" value, even without
        an instrument connection, as the cache values of the :class:`.Platform` object are modified.

        Args:
            parameter (Parameter): Name of the parameter to change.
            value (float | str | bool): New value to set in the parameter.
            alias (str): Alias of the bus where the parameter is set.
            channel_id (int, optional): ID of the channel you want to use to set the parameter. Defaults to None.
        """
        regex_match = re.search(GATE_ALIAS_REGEX, alias)
        if alias == "platform" or parameter == Parameter.DELAY or regex_match is not None:
            if self.digital_compilation_settings is None:
                raise ValueError("Trying to get parameter of gates settings, but no gates settings exist in platform.")
            self.digital_compilation_settings.set_parameter(
                alias=alias, parameter=parameter, value=value, channel_id=channel_id
            )
            return

        element = self.get_element(alias=alias)

        if parameter == Parameter.FLUX:
            self.flux_parameter[alias] = float(value)
            self._process_crosstalk(alias, value)
            self._set_bias_from_element(element)
            return

        element.set_parameter(parameter=parameter, value=value, channel_id=channel_id)

    def _set_bias_from_element(self, element: list[GateEventSettings] | Bus | InstrumentController | Instrument | None):  # type: ignore[union-attr]
        """Sets the right parameter depending on the instrument defined inside the element.
        This is used in the crosstalk correction.
        The instruments included in this function are: QM, QBlox, SPI and QDevil.

        Args:
            element (_type_): runcard elemnt.

        """
        bias = [
            instrument
            for instrument in element.instruments  # type: ignore[union-attr]
            if instrument.name
            in {"QCM", "QRM", "QRM-RF", "QCM-RF", "D5a", "S4g", "quantum_machines_cluster", "qdevil_qdac2"}
        ]

        if len(bias) == 0:
            raise ReferenceError(
                "Flux bus must have one of these instruments:\nQCM, QRM, QRM-RF, QCM-RF, D5a, S4g, quantum_machines_cluster, qdevil_qdac2"
            )
        if len(bias) > 1:
            raise NotImplementedError(
                "Flux bus must not have more than one of these instruments:\nQCM, QRM, QRM-RF, QCM-RF, D5a, S4g, quantum_machines_cluster, qdevil_qdac2"
            )
        if bias[0].name in {"quantum_machines_cluster"}:
            parameter = Parameter.DC_OFFSET
        if bias[0].name in {"D5a", "qdevil_qdac2"}:
            parameter = Parameter.VOLTAGE
        if bias[0].name in {"S4g"}:
            parameter = Parameter.CURRENT
        for flux_alias, flux_value in self.flux_vector.bias_vector.items():  # type: ignore[union-attr]
            flux_element = self.get_element(alias=flux_alias)
            if bias[0].name in {"QCM", "QRM", "QRM-RF", "QCM-RF"}:
                offset_channel = flux_element.instruments[0].awg_sequencers[flux_element.channels[0]].outputs[0]
                if offset_channel == 0:
                    parameter = Parameter.OFFSET_OUT0
                if offset_channel == 1:
                    parameter = Parameter.OFFSET_OUT1
                if offset_channel == 2:
                    parameter = Parameter.OFFSET_OUT2
                if offset_channel == 3:
                    parameter = Parameter.OFFSET_OUT3
            flux_element.set_parameter(parameter=parameter, value=flux_value)
        return

    def _process_crosstalk(self, alias: str, value):
        """Calculates the Current/Voltage of the set parameter based on the value of the flux and the crosstalk matrix"""
        if not self.crosstalk:
            raise ValueError("Crosstalk matrix has not been set")

        bus_list = list(self.crosstalk.matrix.keys())

        if not self.flux_vector:
            self.flux_vector = FluxVector()
        for flux in bus_list:
            if flux not in self.flux_vector.flux_vector.keys():
                self.flux_vector[flux] = 0

        if alias not in self.crosstalk.matrix.keys():
            raise ValueError(f"{alias} not inside crosstalk matrix\n{self.crosstalk}")

        if not self.flux_vector.crosstalk or self.crosstalk != self.flux_vector.crosstalk:
            self.flux_vector.set_crosstalk(self.crosstalk)

        self.flux_vector[alias] = value

    def set_crosstalk(self, crosstalk: CrosstalkMatrix):
        """Sets the crosstalk matrix using the crosstalk matrix class.

        Args:
            crosstalk (CrosstalkMatrix): Crosstalk matrix to implement on the sample
        """
        self.crosstalk = crosstalk

    def set_flux_to_zero(self, bus_list: list[str] | None = None):
        """Set all lines inside the crosstalk / bus list to 0 Phi0, sets everything into the crosstalk offsets.

        Args:
            bus_list (list[str] | None, optional): Optional bus list for all the flux used in the experiment. Defaults to the Crosstalk buses.
        """
        if not self.crosstalk:
            raise ValueError("Crosstalk matrix has not been set")

        if not bus_list:
            bus_list = list(self.crosstalk.matrix.keys())

        for flux_alias in bus_list:
            self.set_parameter(alias=flux_alias, parameter=Parameter.FLUX, value=0)

    def set_bias_to_zero(self, bus_list: list[str] | None = None):
        """Set all lines inside the crosstalk / bus list to 0 Volts / Amperes.

        Args:
            bus_list (list[str] | None, optional): Optional bus list for all the flux used in the experiment. Defaults to the Crosstalk buses.
        """

        if not bus_list:
            if not self.crosstalk:
                raise ValueError("Neither crosstalk matrix nor bus_list has been set")
            bus_list = list(self.crosstalk.matrix.keys())

        if not self.flux_vector:
            self.flux_vector = FluxVector()
        for flux in bus_list:
            if flux not in self.flux_vector.flux_vector.keys():
                self.flux_vector[flux] = 0

        for flux_alias in bus_list:
            element = self.get_element(alias=flux_alias)
            self._set_bias_from_element(element)

    def _load_instruments(self, instruments_dict: list[dict]) -> list[Instrument]:
        """Instantiates all instrument classes from their respective dictionaries.

        Args:
            instruments_dict (list[dict]): List of dictionaries containing the settings of each instrument.

        Returns:
            list[Instrument]: List of instantiated instrument classes.
        """
        instruments = []
        for instrument in instruments_dict:
            local_dict = deepcopy(instrument)
            instruments.append(InstrumentFactory.get(local_dict.pop(RUNCARD.NAME))(settings=local_dict))
        return instruments

    def _load_instrument_controllers(self, instrument_controllers_dict: list[dict]) -> list[InstrumentController]:
        """Instantiates all instrument controller classes from their respective dictionaries.

        Args:
            instrument_controllers_dict (list[dict]): List of dictionaries containing
            the settings of each instrument controller.

        Returns:
            list[InstrumentController]: List of instantiated instrument controller classes.
        """
        instrument_controllers = []
        for instrument_controller in instrument_controllers_dict:
            local_dict = deepcopy(instrument_controller)
            instrument_controllers.append(
                InstrumentControllerFactory.get(local_dict.pop(RUNCARD.NAME))(
                    settings=local_dict, loaded_instruments=self.instruments
                )
            )
        return instrument_controllers

    def to_dict(self):
        """Returns all platform information as a dictionary, called the :ref:`runcard <runcards>`. Used for the platform serialization.

        Returns:
            dict: Dictionary of the serialized platform
        """
        name_dict = {RUNCARD.NAME: self.name}
        instrument_dict = {RUNCARD.INSTRUMENTS: self.instruments.to_dict()}
        instrument_controllers_dict = {RUNCARD.INSTRUMENT_CONTROLLERS: self.instrument_controllers.to_dict()}
        buses_dict = {RUNCARD.BUSES: self.buses.to_dict()}
        digital_dict = {
            RUNCARD.DIGITAL: (
                self.digital_compilation_settings.to_dict() if self.digital_compilation_settings is not None else None
            )
        }
        analog_dict = {
            RUNCARD.ANALOG: (
                self.analog_compilation_settings.to_dict() if self.analog_compilation_settings is not None else None
            )
        }

        return name_dict | instrument_dict | instrument_controllers_dict | buses_dict | digital_dict | analog_dict

    def __str__(self) -> str:
        """String representation of the platform.

        Returns:
            str: Name of the platform.
        """
        return str(YAML(typ="safe").dump(self.to_dict(), io.BytesIO()))

    @contextmanager
    def session(self):
        """Context manager to manage platform session, ensuring that resources are always released."""
        cleanup_methods = []
        cleanup_errors = []
        try:
            # Track successfully called setup methods and their cleanup counterparts
            self.connect()
            cleanup_methods.append(self.disconnect)  # Store disconnect for cleanup

            self.initial_setup()  # No specific cleanup for initial_setup

            self.turn_on_instruments()
            cleanup_methods.append(self.turn_off_instruments)  # Store turn_off_instruments for cleanup

            yield  # Experiment logic goes here

        except Exception as e:
            logger.error(f"An error occurred: {e}")
            raise  # Re-raise the exception for further handling
        finally:
            # Call the cleanup methods in reverse order
            for cleanup_method in reversed(cleanup_methods):
                try:
                    cleanup_method()
                except Exception as e:  # noqa: BLE001
                    logger.error(f"Error during cleanup: {e}")
                    cleanup_errors.append(e)

            # Raise any exception that might have happened during cleanup
            if cleanup_errors:
                raise ExceptionGroup("Exceptions occurred during cleanup", cleanup_errors)

    def compile_annealing_program(
        self,
        annealing_program_dict: list[dict[str, dict[str, float]]],
        transpiler: Callable,
        calibration: Calibration,
        num_averages: int = 1000,
        num_shots: int = 1,
        preparation_block: str = "preparation",
        measurement_block: str = "measurement",
    ) -> QProgram:
        """
        Compile an annealing program into a `QProgram` by mapping Ising coefficients to flux waveforms.

        This method takes an annealing program, represented as a time-ordered list of circuit elements with
        corresponding Ising coefficients, and compiles it into a quantum program (QProgram) that can be
        executed on a quantum annealing hardware setup.

        The input `annealing_program_dict` is structured as a list of dictionaries, each representing a
        specific time point. Each dictionary maps qubit and coupler identifiers to Ising terms (`sigma_x`,
        `sigma_y`, `sigma_z`), indicating the coefficient values for each term. For example:

        .. code-block:: python

            [
                {"qubit_0": {"sigma_x": 0, "sigma_y": 1, "sigma_z": 2}, "coupler_1_0": {...}},
                {...},  # time=1ns
                ...,
            ]

        Using the provided `transpiler`, these Ising coefficients are converted to flux values. These fluxes
        are then transformed into waveforms assigned to specific hardware buses, as defined by a `flux_to_bus`
        mapping in the analog compilation settings.

        Args:
            annealing_program_dict (list[dict[str, dict[str, float]]]): The time-ordered list of qubit and
                coupler Ising coefficients to be compiled.
            transpiler (Callable): A function to convert Ising parameters (delta, epsilon) to flux values
                (phix, phiz).
            calibration (Calibration): Calibration data containing the required blocks (e.g., `preparation`,
                `measurement`) and any applicable crosstalk corrections.
            num_averages (int, optional): Number of times the program should be averaged per shot. Defaults to 1000.
            num_shots (int, optional): Number of shots to execute the program. Defaults to 1.
            preparation_block (str, optional): Name of the calibration block used for preparation. Defaults to "preparation".
            measurement_block (str, optional): Name of the calibration block used for measurement. Defaults to "measurement".

        Returns:
            QProgram: A compiled quantum program (QProgram) ready for execution on the target hardware.

        Raises:
            ValueError: If the flux-to-bus topology is not defined in the analog compilation settings.
            ValueError: If the specified `measurement_block` is not available in the calibration.

        Notes:
            - The method checks for essential compilation settings and calibrated blocks, ensuring the program can be executed successfully.
            - Transpiled waveforms are adjusted for crosstalk when a crosstalk matrix is available in the calibration.
            - Execution includes optional `preparation_block` and synchronizes waveforms before the final `measurement_block`.

        """
        if self.analog_compilation_settings is None:
            raise ValueError("Flux to bus topology not given in the runcard")

        if not calibration.has_block(name=measurement_block):
            raise ValueError("The calibrated measurement is not present in the calibration file.")

        annealing_program = AnnealingProgram(
            flux_to_bus_topology=self.analog_compilation_settings.flux_control_topology,
            annealing_program=annealing_program_dict,
        )
        annealing_program.transpile(transpiler)
        crosstalk_matrix = calibration.crosstalk_matrix.inverse() if calibration.crosstalk_matrix is not None else None
        annealing_waveforms = annealing_program.get_waveforms(crosstalk_matrix=crosstalk_matrix, minimum_clock_time=4)

        qp_annealing = QProgram()
        shots_variable = qp_annealing.variable("num_shots", Domain.Scalar, int)

        with qp_annealing.for_loop(variable=shots_variable, start=0, stop=num_shots, step=1):
            with qp_annealing.average(num_averages):
                if calibration.has_block(name=preparation_block):
                    qp_annealing.insert_block(calibration.get_block(name=preparation_block))
                    qp_annealing.sync()
                for bus, waveform in annealing_waveforms.items():
                    qp_annealing.play(bus=bus, waveform=waveform)
                qp_annealing.sync()
                qp_annealing.insert_block(calibration.get_block(name=measurement_block))

        return qp_annealing

    def execute_annealing_program(
        self,
        annealing_program_dict: list[dict[str, dict[str, float]]],
        transpiler: Callable,
        calibration: Calibration,
        num_averages: int = 1000,
        num_shots: int = 1,
        preparation_block: str = "preparation",
        measurement_block: str = "measurement",
        bus_mapping: dict[str, str] | None = None,
        debug: bool = False,
    ) -> QProgramResults:
        """Given an annealing program execute it as a qprogram.
        The annealing program should contain a time ordered list of circuit elements and their corresponging ising coefficients as a dictionary. Example structure:

        .. code-block:: python

            [
                {"qubit_0": {"sigma_x" : 0, "sigma_y" : 1, "sigma_z" : 2},
                "coupler_1_0 : {...},
                },      # time=0ns
                {...},  # time=1ns
            .
            .
            .
            ]

        This dictionary containing ising coefficients is transpiled to fluxes using the given transpiler. Then the correspoinding waveforms are obtained and assigned to a bus
        from the bus to flux mapping given by the runcard.

        Args:
            annealing_program_dict (list[dict[str, dict[str, float]]]): annealing program to run
            transpiler (Callable): ising to flux transpiler. The transpiler should take 2 values as arguments (delta, epsilon) and return 2 values (phix, phiz)
            averages (int, optional): Amount of times to run and average the program over. Defaults to 1.
            debug (bool, optional): Whether to create debug information. For ``Qblox`` clusters all the program information is printed on screen.
                For ``Quantum Machines`` clusters a ``.py`` file is created containing the ``QUA`` and config compilation. Defaults to False.
        """

        qprogram = self.compile_annealing_program(
            annealing_program_dict=annealing_program_dict,
            transpiler=transpiler,
            calibration=calibration,
            num_averages=num_averages,
            num_shots=num_shots,
            preparation_block=preparation_block,
            measurement_block=measurement_block,
        )
        return self.execute_qprogram(qprogram=qprogram, calibration=calibration, bus_mapping=bus_mapping, debug=debug)

    def execute_experiment(self, experiment: Experiment, calibration: Calibration | None = None) -> str:
        """Executes a quantum experiment on the platform.

        This method manages the execution of a given `Experiment` on the platform by utilizing an `ExperimentExecutor`. It orchestrates the entire process, including traversing the experiment's structure, handling loops and operations, and streaming results in real-time to ensure data integrity. The results are saved in a timestamped directory within the specified `base_data_path`.

        Args:
            experiment (Experiment): The experiment object defining the sequence of operations and loops.

        Returns:
            str: The path to the file where the results are stored.

        Example:
            .. code-block:: python

                from qililab import Experiment

                # Initialize your experiment
                experiment = Experiment(label="my_experiment")
                # Add variables, loops, and operations to the experiment
                # ...

                # Define the base path for storing experiment results
                platform.experiment_results_base_path = "/data/experiments"

                # Execute the experiment on the platform
                results_path = platform.execute_experiment(experiment=experiment)
                print(f"Results saved to {results_path}")

        Note:
            - Ensure that the experiment is properly configured before execution.
            - The results will be saved in a directory within the `experiment_results_base_path` according to the `platform.experiment_results_path_format`. The default format is `{date}/{time}/{label}.h5`.
            - This method handles the setup and execution internally, providing a simplified interface for experiment execution.
        """
        executor = ExperimentExecutor(platform=self, experiment=experiment)
        return executor.execute()

    def compile_qprogram(
        self, qprogram: QProgram, bus_mapping: dict[str, str] | None = None, calibration: Calibration | None = None
    ) -> QbloxCompilationOutput | QuantumMachinesCompilationOutput:
        bus_aliases = {bus_mapping[bus] if bus_mapping and bus in bus_mapping else bus for bus in qprogram.buses}
        buses = [self.buses.get(alias=bus_alias) for bus_alias in bus_aliases]
        instruments = {
            instrument
            for bus in buses
            for instrument in bus.instruments
            if isinstance(instrument, (QbloxModule, QuantumMachinesCluster))
        }
        if all(isinstance(instrument, QbloxModule) for instrument in instruments):
            # Retrieve the time of flight parameter from settings
            times_of_flight = {
                bus.alias: int(bus.get_parameter(Parameter.TIME_OF_FLIGHT)) for bus in buses if bus.has_adc()
            }
            delays = {bus.alias: int(bus.get_parameter(Parameter.DELAY)) for bus in buses}
            # Determine what should be the initial value of the markers for each bus.
            # This depends on the model of the associated Qblox module and the `output` setting of the associated sequencer.
            markers = {}
            for bus in buses:
                for instrument, channel in zip(bus.instruments, bus.channels):
                    if isinstance(instrument, QbloxModule):
                        sequencer = instrument.get_sequencer(sequencer_id=channel)
                        if instrument.name == InstrumentName.QCMRF:
                            markers[bus.alias] = "".join(
                                ["1" if i in [0, 1] and i in sequencer.outputs else "0" for i in range(4)]
                            )[::-1]
                        elif instrument.name == InstrumentName.QRMRF:
                            markers[bus.alias] = "".join(
                                ["1" if i in [1] and i - 1 in sequencer.outputs else "0" for i in range(4)]
                            )[::-1]
                        else:
                            markers[bus.alias] = "0000"
            qblox_compiler = QbloxCompiler()
            return qblox_compiler.compile(
                qprogram=qprogram,
                bus_mapping=bus_mapping,
                calibration=calibration,
                times_of_flight=times_of_flight,
                delays=delays,
                markers=markers,
            )
        if all(isinstance(instrument, QuantumMachinesCluster) for instrument in instruments):
            if len(instruments) != 1:
                raise NotImplementedError(
                    "Executing QProgram in more than one Quantum Machines Cluster is not supported."
                )
            thresholds: dict[str, float] = {
                bus.alias: float(bus.get_parameter(parameter=Parameter.THRESHOLD) or 0.0)
                for bus in buses
                if bus.has_adc()
            }
            threshold_rotations: dict[str, float] = {
                bus.alias: float(bus.get_parameter(parameter=Parameter.THRESHOLD_ROTATION) or 0.0)
                for bus in buses
                if bus.has_adc()
            }

            compiler = QuantumMachinesCompiler()
            return compiler.compile(
                qprogram=qprogram,
                bus_mapping=bus_mapping,
                thresholds=thresholds,
                threshold_rotations=threshold_rotations,
                calibration=calibration,
            )
        raise NotImplementedError("Compiling QProgram for a mixture of instruments is not supported.")

    def execute_compilation_output(
        self,
        output: QbloxCompilationOutput | QuantumMachinesCompilationOutput,
        debug: bool = False,
    ):
        if isinstance(output, QbloxCompilationOutput):
            return self._execute_qblox_compilation_output(output=output, debug=debug)

        buses = [self.buses.get(alias=bus_alias) for bus_alias in output.qprogram.buses]
        instruments = {instrument for bus in buses for instrument in bus.instruments if bus.has_adc()}
        if len(instruments) != 1:
            raise NotImplementedError("Executing QProgram in more than one Quantum Machines Cluster is not supported.")
        cluster: QuantumMachinesCluster = cast("QuantumMachinesCluster", next(iter(instruments)))
        return self._execute_quantum_machines_compilation_output(output=output, cluster=cluster, debug=debug)

    def _execute_qblox_compilation_output(self, output: QbloxCompilationOutput, debug: bool = False):
        sequences, acquisitions = output.sequences, output.acquisitions
        buses = {bus_alias: self.buses.get(alias=bus_alias) for bus_alias in sequences}
        for bus_alias, bus in buses.items():
            if bus.distortions:
                for distortion in bus.distortions:
                    for waveform in sequences[bus_alias]._waveforms._waveforms:
                        sequences[bus_alias]._waveforms.modify(waveform.name, distortion.apply(waveform.data))
        if debug:
            with open("debug_qblox_execution.txt", "w", encoding="utf-8") as sourceFile:
                for bus_alias, sequence in sequences.items():
                    print(f"Bus {bus_alias}:", file=sourceFile)
                    print(str(sequence._program), file=sourceFile)
                    print(file=sourceFile)

        # Upload sequences
        for bus_alias in sequences:
            sequence_hash = hash_qpy_sequence(sequence=sequences[bus_alias])
            if bus_alias not in self._qpy_sequence_cache or self._qpy_sequence_cache[bus_alias] != sequence_hash:
                buses[bus_alias].upload_qpysequence(qpysequence=sequences[bus_alias])
                self._qpy_sequence_cache[bus_alias] = sequence_hash
            # sync all relevant sequences
            for instrument, channel in zip(buses[bus_alias].instruments, buses[bus_alias].channels):
                if isinstance(instrument, QbloxModule):
                    instrument.sync_sequencer(sequencer_id=int(channel))

        # Execute sequences
        for bus_alias in sequences:
            buses[bus_alias].run()

        # Acquire results
        results = QProgramResults()
        for bus_alias, bus in buses.items():
            if bus.has_adc():
                for instrument, channel in zip(buses[bus_alias].instruments, buses[bus_alias].channels):
                    if isinstance(instrument, QbloxModule):
                        bus_results = bus.acquire_qprogram_results(
                            acquisitions=acquisitions[bus_alias], channel_id=int(channel)
                        )
                        for bus_result in bus_results:
                            results.append_result(bus=bus_alias, result=bus_result)

        # Reset instrument settings
        for bus_alias in sequences:
            for instrument, channel in zip(buses[bus_alias].instruments, buses[bus_alias].channels):
                if isinstance(instrument, QbloxModule):
                    instrument.desync_sequencer(sequencer_id=int(channel))

        return results

    def _execute_quantum_machines_compilation_output(
        self,
        output: QuantumMachinesCompilationOutput,
        cluster: QuantumMachinesCluster,
        debug: bool = False,
    ):
        qua, configuration, measurements = output.qua, output.configuration, output.measurements
        try:
            cluster.append_configuration(configuration=configuration)

            if debug:
                with open("debug_qm_execution.py", "w", encoding="utf-8") as sourceFile:
                    print(generate_qua_script(qua, cluster.config), file=sourceFile)

            compiled_program_id = cluster.compile(program=qua)
            job = cluster.run_compiled_program(compiled_program_id=compiled_program_id)

            acquisitions = cluster.get_acquisitions(job=job)

            results = QProgramResults()
            # Doing manual classification of results as QM does not return thresholded values like Qblox
            for measurement in measurements:
                measurement_result = QuantumMachinesMeasurementResult(
                    measurement.bus,
                    *[acquisitions[handle] for handle in measurement.result_handles],
                )
                measurement_result.set_classification_threshold(measurement.threshold)
                results.append_result(bus=measurement.bus, result=measurement_result)

            return results

        except Exception as e:
            cluster.turn_off()
            raise e

        return results

    def execute_qprogram(
        self,
        qprogram: QProgram,
        bus_mapping: dict[str, str] | None = None,
        calibration: Calibration | None = None,
        debug: bool = False,
    ) -> QProgramResults:
        """Execute a :class:`.QProgram` using the platform instruments.

        |

        **The execution is done in the following steps:**

        1. Compile the QProgram.
        2. Run the compiled QProgram.
        3. Acquire the results.

        |

        **The execution can be done for (buses associated to) two different type of clusters:**

        - For ``Qblox`` modules, the compilation is done using the :class:`.QbloxCompiler`. Which compiles the :class:`.QProgram` into``Q1ASM`` for multiple sequencers based on each bus, uploads and executes the sequences, and acquires the results.
        - For ``Quantum Machines`` clusters, the compilation is done using the :class:`.QuantumMachinesCompiler`. This compiler transforms the :class:`.QProgram` into ``QUA``, the programming language of ``Quantum Machines`` hardware. It then executes the resulting ``QUA`` program and returns the results, organized by bus.

        Args:
            qprogram (QProgram): The :class:`.QProgram` to execute.
            bus_mapping (dict[str, str], optional): A dictionary mapping the buses in the :class:`.QProgram` (keys )to the buses in the platform (values).
                It is useful for mapping a generic :class:`.QProgram` to a specific experiment. Defaults to None.
            calibration (Calibration, optional): :class:`.Calibration` instance containing information of previously calibrated values, like waveforms, weights and crosstalk matrix. Defaults to None.
            debug (bool, optional): Whether to create debug information. For ``Qblox`` clusters all the program information is printed on screen.
                For ``Quantum Machines`` clusters a ``.py`` file is created containing the ``QUA`` and config compilation. Defaults to False.

        Returns:
            QProgramResults: The results of the execution. ``QProgramResults.results()`` returns a dictionary (``dict[str, list[Result]]``) of measurement results.
            The keys correspond to the buses a measurement were performed upon, and the values are the list of measurement results in chronological order.
        """
        output = self.compile_qprogram(qprogram=qprogram, bus_mapping=bus_mapping, calibration=calibration)
        return self.execute_compilation_output(output=output, debug=debug)

    def execute_qprograms_parallel(
        self,
        qprograms: list[QProgram],
        bus_mapping: dict[str, str] | None = None,
        calibration: Calibration | None = None,
        debug: bool = False,
    ) -> list[QProgramResults]:
        """Compiles a list of qprograms to be executed in parallel. Then it calls the execute_compilation_outputs_parallel method to execute the compiled qprograms.
        It loads each qprogram into a different sequencer and uses the multiplexing capabilities of QBlox to run all sequencers at the same time.

        **The execution can be done for (buses associated to) Qblox only. And it can only be done for qprograms that do not share buses.**

        Args:
            qprograms (QProgram): A list of the :class:`.QProgram` to execute.
            bus_mapping (dict[str, str], optional): A dictionary mapping the buses in the :class:`.QProgram` (keys )to the buses in the platform (values).
                It is useful for mapping a generic :class:`.QProgram` to a specific experiment. Defaults to None.
            calibration (Calibration, optional): :class:`.Calibration` instance containing information of previously calibrated values, like waveforms, weights and crosstalk matrix. Defaults to None.
            debug (bool, optional): Whether to create debug information. For ``Qblox`` clusters all the program information is printed on screen.
                Defaults to False.

        Returns:
            QProgramResults: The results of the execution. ``QProgramResults.results()`` returns a list of dictionary (``dict[str, list[Result]]``) of measurement results.
            Each element of the list corresponds to a sequencer.
            The keys correspond to the buses a measurement were performed upon, and the values are the list of measurement results in chronological order.
        """
        buses_per_qprogram = [qprogram.buses for qprogram in qprograms]
        total_buses = sum(len(s) for s in buses_per_qprogram)
        unique_buses = len(set.union(*buses_per_qprogram))
        if total_buses != unique_buses:
            raise ValueError("QPrograms cannot be executed in parallel.")
        outputs = [
            self.compile_qprogram(qprogram=qprogram, bus_mapping=bus_mapping, calibration=calibration)
            for qprogram in qprograms
        ]
        if any(isinstance(output, QuantumMachinesCompilationOutput) for output in outputs):
            raise ValueError("Parallel execution is not supported in Quantum Machines.")
        return self.execute_compilation_outputs_parallel(
            outputs=cast("list[QbloxCompilationOutput]", outputs), debug=debug
        )

    def execute_compilation_outputs_parallel(
        self,
        outputs: list[QbloxCompilationOutput],
        debug: bool = False,
    ):
        """Execute compiled qprograms in parallel.
        It loads each qprogram into a different sequencer and uses the multiplexing capabilities of QBlox to run all sequencers at the same time.

        |

        **The execution is done in the following steps:**

        1. Upload all sequences.
        2. Execute all sequences.
        3. Acquire the results.

        |

        **The execution can be done for (buses associated to) Qblox only.**

        Args:
            outputs: A list of the compiled qprograms.
            debug (bool, optional): Whether to create debug information. For ``Qblox`` clusters all the program information is printed on screen.
                Defaults to False.

        Returns:
            QProgramResults: The results of the execution. ``QProgramResults.results()`` returns a list of dictionary (``dict[str, list[Result]]``) of measurement results.
            Each element of the list corresponds to a sequencer.
            The keys correspond to the buses a measurement were performed upon, and the values are the list of measurement results in chronological order.
        """
        sequences_per_qprogram = [output.sequences for output in outputs]
        aquisitions_per_qprogram = [output.acquisitions for output in outputs]
        buses_per_qprogram = [
            {bus_alias: self.buses.get(alias=bus_alias) for bus_alias in sequences}
            for sequences in sequences_per_qprogram
        ]
        for qprogram_idx, buses in enumerate(buses_per_qprogram):
            for bus_alias, bus in buses.items():
                if bus.distortions:
                    for distortion in bus.distortions:
                        for waveform in sequences_per_qprogram[qprogram_idx][bus_alias]._waveforms._waveforms:
                            sequences_per_qprogram[qprogram_idx][bus_alias]._waveforms.modify(
                                waveform.name, distortion.apply(waveform.data)
                            )

        if debug:
            with open("debug_qblox_execution.txt", "w", encoding="utf-8") as sourceFile:
                for qprogram_idx, sequences in enumerate(sequences_per_qprogram):
                    print(f"QProgram {qprogram_idx}:", file=sourceFile)
                    for bus_alias, sequence in sequences.items():
                        print(f"Bus {bus_alias}:", file=sourceFile)
                        print(str(sequence._program), file=sourceFile)
                        print(file=sourceFile)

        # Upload sequences
        for qprogram_idx, sequences in enumerate(sequences_per_qprogram):
            for bus_alias in sequences:
                sequence_hash = hash_qpy_sequence(sequence=sequences[bus_alias])
                if bus_alias not in self._qpy_sequence_cache or self._qpy_sequence_cache[bus_alias] != sequence_hash:
                    buses_per_qprogram[qprogram_idx][bus_alias].upload_qpysequence(qpysequence=sequences[bus_alias])
                    self._qpy_sequence_cache[bus_alias] = sequence_hash
                # sync all relevant sequences
                for instrument, channel in zip(
                    buses_per_qprogram[qprogram_idx][bus_alias].instruments,
                    buses_per_qprogram[qprogram_idx][bus_alias].channels,
                ):
                    if isinstance(instrument, QbloxModule):
                        instrument.sync_sequencer(sequencer_id=int(channel))

        # Execute sequences
        for qprogram_idx, sequences in enumerate(sequences_per_qprogram):
            for bus_alias in sequences:
                buses_per_qprogram[qprogram_idx][bus_alias].run()

        # Acquire results
        results = [QProgramResults() for _ in outputs]
        for qprogram_idx, buses in enumerate(buses_per_qprogram):
            for bus_alias, bus in buses.items():
                if bus.has_adc():
                    for instrument, channel in zip(buses[bus_alias].instruments, buses[bus_alias].channels):
                        if isinstance(instrument, QbloxModule):
                            bus_results = bus.acquire_qprogram_results(
                                acquisitions=aquisitions_per_qprogram[qprogram_idx][bus_alias], channel_id=int(channel)
                            )
                            for bus_result in bus_results:
                                results[qprogram_idx].append_result(bus=bus_alias, result=bus_result)

        # Reset instrument settings
        for qprogram_idx, sequences in enumerate(sequences_per_qprogram):
            for bus_alias in sequences:
                for instrument, channel in zip(
                    buses_per_qprogram[qprogram_idx][bus_alias].instruments,
                    buses_per_qprogram[qprogram_idx][bus_alias].channels,
                ):
                    if isinstance(instrument, QbloxModule):
                        instrument.desync_sequencer(sequencer_id=int(channel))

        return results

    def execute(
        self,
        program: PulseSchedule | Circuit,
        num_avg: int,
        repetition_duration: int = 200_000,
        num_bins: int = 1,
        queue: Queue | None = None,
        transpilation_config: DigitalTranspilationConfig | None = None,
    ) -> Result | QbloxResult:
        """Compiles and executes a circuit or a pulse schedule, using the platform instruments.

        If the ``program`` argument is a :class:`.Circuit`, it will first be translated into a :class:`.PulseSchedule` using the transpilation
        settings of the platform and the passed transpile onfiguration. Then the pulse schedules will be compiled into the assembly programs and executed.

        To compile to assembly programs, the ``platform.compile()`` method is called; check its documentation for more information.

        The transpilation is performed using the :meth:`.CircuitTranspiler.transpile_circuit()` method. Refer to the method's documentation or :ref:`Transpilation <transpilation>` for more detailed information.

        The main stages of this process are: **1.** Routing, **2.** Canceling Hermitian pairs, **3.** Translate to native gates, **4.** Correcting Drag phases, **5** Optimize Drag gates, **6.** Convert to pulse schedule.

        .. note ::

            Default steps are only: **3.**, **4.**, and **6.**, since they are always needed.

            To do Step **1.** set routing=True in transpilation_config (default behavior skips it).

            To do Steps **2.** and **5.** set optimize=True in transpilation_config (default behavior skips it)

        Args:
            program (``PulseSchedule`` | ``Circuit``): Circuit or pulse schedule to execute.
            num_avg (int): Number of hardware averages used.
            repetition_duration (int): Minimum duration of a single execution. Defaults to 200_000.
            num_bins (int, optional): Number of bins used. Defaults to 1.
            queue (Queue, optional): External queue used for asynchronous data handling. Defaults to None.
            transpilation_config (DigitalTranspilationConfig, optional): :class:`.DigitalTranspilationConfig` dataclass containing
                the configuration used during transpilation. Defaults to ``None`` (not changing any default value).
                Check the class:`.DigitalTranspilationConfig` documentation for the keys and values it can contain.

        Returns:
            Result: Result obtained from the execution. This corresponds to a numpy array that depending on the
                platform configuration may contain the following:

                - Scope acquisition is enabled: An array with dimension `(2, N)` which contain the scope data for
                    path0 (I) and path1 (Q). N corresponds to the length of the scope measured.

                - Scope acquisition disabled: An array with dimension `(#sequencers, 2, #bins)`.

        |

        Example Usage:

        .. code-block:: python

            from qibo import gates, Circuit
            from qibo.transpiler import ReverseTraversal, Sabre
            from qililab import build_platform
            from qililab.digital import DigitalTranspilationConfig

            # Create circuit:
            c = Circuit(5)
            c.add(gates.CNOT(1, 0))

            # Create platform:
            platform = build_platform(runcard="<path_to_runcard>")
            transp_config = DigitalTranspilationConfig(routing=True, optimize=False, router=Sabre, placer=ReverseTraversal)

            # Execute with automatic transpilation:
            result = platform.execute(c, num_avg=1000, transpilation_config=transp_config)
        """
        # Compile pulse schedule
        programs, final_layout = self.compile(program, num_avg, repetition_duration, num_bins, transpilation_config)

        # Upload pulse schedule
        for bus_alias in programs:
            bus = self.buses.get(alias=bus_alias)
            bus.upload()

        # Execute pulse schedule
        for bus_alias in programs:
            bus = self.buses.get(alias=bus_alias)
            bus.run()

        # Acquire results
        readout_buses = [bus for bus in self.buses if bus.alias in programs and bus.has_adc()]
        results: list[Result] = []
        for bus in readout_buses:
            result = bus.acquire_result()
            if queue is not None:
                queue.put_nowait(item=result)
            if not np.all(np.isnan(result.array)):
                results.append(result)

        for instrument in self.instruments.elements:
            if isinstance(instrument, QbloxModule):
                instrument.desync_sequencers()

        # Flatten results if more than one readout bus was used for a qblox module
        if len(results) > 1:
            result = QbloxResult(
                integration_lengths=[length for result in results for length in result.integration_lengths],  # type: ignore[attr-defined]
                qblox_raw_results=[raw_result for result in results for raw_result in result.qblox_raw_results],  # type: ignore[attr-defined]
            )
        elif not results:
            raise ValueError("There are no readout buses in the platform.")
        else:
            result = results[0]

        if isinstance(program, Circuit):
            result = self._order_result(result, program, final_layout)

        return result

    @staticmethod
    def _order_result(result: Result, circuit: Circuit, final_layout: list[int] | None) -> Result:
        """Order the results of the execution as they are ordered in the input circuit.

        Finds the absolute order of each measurement for each qubit and its corresponding key in the
        same format as in qblox's aqcuisitions dictionary (#qubit, #qubit_measurement).

        Then it orders results in the same measurement order as the one in circuit.queue.

        Args:
            result (Result): Result obtained from the execution
            circuit (Circuit): Qibo circuit being executed
            final_layouts (list[int], optional): Final layout of the original logical qubits in the physical circuit:
                [Logical qubit in wire 1, Logical qubit in wire 2, ...] (None = trivial mapping).

        Returns:
            Result: Result obtained from the execution, with each measurement in the same order as in circuit.queue.
        """
        if not isinstance(result, QbloxResult):
            raise NotImplementedError("Result ordering is only implemented for qblox results")

        # register the overall order of all qubit measurements.
        qubits_m = {}
        order = {}
        # iterate over qubits measured in same order as they appear in the circuit
        # TODO: You need to check where each measurement is, since SWAPs can be after a measurement...
        # FIXME: In the meanwhile do it asuming the Measurement is the last gate for each qubit
        for i, qubit in enumerate(qubit for gate in circuit.queue for qubit in gate.qubits if isinstance(gate, M)):
            if qubit not in qubits_m:
                qubits_m[qubit] = 0
            order[qubit, qubits_m[qubit]] = i
            qubits_m[qubit] += 1
        if len(order) != len(result.qblox_raw_results):
            raise ValueError(
                f"Number of measurements in the circuit {len(order)} does not match number of acquisitions {len(result.qblox_raw_results)}"
            )

        # Tell users that the final layout is being undone:
        logger.info(
            "Undoing final physical qubit mapping, so you get back the original qubit order in your logical circuit."
        )

        # allocate each measurement its corresponding index in the results list
        results = [None] * len(order)  # type: list | list[dict]
        for qblox_result in result.qblox_raw_results:
            measurement = qblox_result["measurement"]
            physical_qubit = qblox_result["qubit"]
            original_logical_qubit = final_layout[physical_qubit] if final_layout else physical_qubit

            # TODO:Check this is correct, or how it works with multiple measurements per qubit:
            original_measure_num = order[original_logical_qubit, measurement]
            results[original_measure_num] = qblox_result

        return QbloxResult(integration_lengths=result.integration_lengths, qblox_raw_results=results)

    def compile(
        self,
        program: PulseSchedule | Circuit,
        num_avg: int,
        repetition_duration: int,
        num_bins: int,
        transpilation_config: DigitalTranspilationConfig | None = None,
    ) -> tuple[dict[str, list[QpySequence]], list[int] | None]:
        """Compiles the circuit / pulse schedule into a set of assembly programs, to be uploaded into the awg buses.

        If the ``program`` argument is a :class:`.Circuit`, it will first be translated into a :class:`.PulseSchedule` using the transpilation
        settings of the platform and passed  transpile configuration. Then the pulse schedules will be compiled into the assembly programs.

        .. note::

            Compile is called during ``platform.execute()``, check its documentation for more information.

        The transpilation is performed using the :meth:`.CircuitTranspiler.transpile_circuit()` method. Refer to the method's documentation or :ref:`Transpilation <transpilation>` for more detailed information.

        The main stages of this process are: **1.** Routing, **2.** Canceling Hermitian pairs, **3.** Translate to native gates, **4.** Correcting Drag phases, **5** Optimize Drag gates, **6.** Convert to pulse schedule.

        .. note ::

            Default steps are only: **3.**, **4.**, and **6.**, since they are always needed.

            To do Step **1.** set routing=True in transpilation_config (default behavior skips it).

            To do Steps **2.** and **5.** set optimize=True in transpilation_config (default behavior skips it)

        Args:
            program (PulseSchedule | Circuit): Circuit or pulse schedule to compile.
            num_avg (int): Number of hardware averages used.
            repetition_duration (int): Minimum duration of a single execution.
            num_bins (int): Number of bins used.
            transpilation_config (DigitalTranspilationConfig, optional): :class:`.DigitalTranspilationConfig` dataclass containing
                the configuration used during transpilation. Defaults to ``None`` (not changing any default value).
                Check the class:`.DigitalTranspilationConfig` documentation for the keys and values it can contain.

        Returns:
            tuple[dict, list[int] | None]: Tuple containing the dictionary of compiled assembly programs (The key is the bus alias (``str``),
                and the value is the assembly compilation (``list``)), and its corresponding final layout (Initial Re-mapping + SWAPs routing) of
                the Original Logical Qubits (l_q) in the physical circuit (wires): [l_q in wire 0, l_q in wire 1, ...] (None = trivial mapping).

        Raises:
            ValueError: raises value error if the circuit execution time is longer than ``repetition_duration`` for some qubit.
        """
        # We have a circular import because Platform uses CircuitToPulses and vice versa
        if self.digital_compilation_settings is None:
            raise ValueError("Cannot compile Qibo Circuit or Pulse Schedule without gates settings.")

        if isinstance(program, Circuit):
            transpiler = CircuitTranspiler(settings=self.digital_compilation_settings)
            pulse_schedule, final_layout = transpiler.transpile_circuit(program, transpilation_config)

        elif isinstance(program, PulseSchedule):
            pulse_schedule = program
            final_layout = None

        else:
            raise ValueError(
                f"Program to execute can only be either a single circuit or a pulse schedule. Got program of type {type(program)} instead"
            )

        module_and_sequencer_per_bus: dict[str, ModuleSequencer] = {
            element.bus_alias: ModuleSequencer(module=instrument, sequencer=instrument.get_sequencer(channel))
            for element in pulse_schedule.elements
            for instrument, channel in zip(
                self.buses.get(alias=element.bus_alias).instruments, self.buses.get(alias=element.bus_alias).channels
            )
            if isinstance(instrument, QbloxModule)
        }

        compiler = PulseQbloxCompiler(
            buses=self.digital_compilation_settings.buses,
            module_and_sequencer_per_bus=module_and_sequencer_per_bus,
        )

        compiled_programs = compiler.compile(
            pulse_schedule=pulse_schedule, num_avg=num_avg, repetition_duration=repetition_duration, num_bins=num_bins
        )

        return compiled_programs, final_layout

    def calibrate_mixers(self, alias: str, cal_type: str, channel_id: ChannelID | None = None):
        bus = self.get_element(alias=alias)
        for instrument, instrument_channel in zip(bus.instruments, bus.channels):
            if instrument.name == InstrumentName.QRMRF:
                if channel_id is not None and channel_id == instrument_channel:
                    instrument.calibrate_mixers(cal_type, instrument_channel)
            elif instrument.name == InstrumentName.QCMRF:
                if channel_id is not None and channel_id == instrument_channel:
                    instrument.calibrate_mixers(cal_type, channel_id)
            else:
                raise AttributeError("Mixers calibration not implemented for this instrument.")

    def draw(self, qprogram: QProgram, averages_displayed: bool = False):
        """Draw the QProgram using QBlox Compiler

        Args:
            averages_displayed (bool): False means that all loops on the sequencer starting with avg will only loop once, and True shows all iterations.
                                        The default is False.
        """

        runcard_data = self._data_draw()
        qblox_draw = QbloxDraw()
        sequencer = self.compile_qprogram(qprogram)
        result = qblox_draw.draw(sequencer, runcard_data, averages_displayed)

        return result



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Rectangular pulse shape."""

from dataclasses import dataclass

import numpy as np

from qililab.pulse.pulse_shape.pulse_shape import PulseShape
from qililab.typings import PulseShapeName
from qililab.utils import Factory


@Factory.register
@dataclass(frozen=True, eq=True)
class Rectangular(PulseShape):
    """Rectangular/square pulse shape. Given by a constant height line.

    Examples:
        To get the envelope of a rectangular shape, with ``amplitude`` equal to ``X``, you need to do:

        .. code-block:: python

            from qililab.pulse.pulse_shape import Rectangular

            rectangular_envelope = Rectangular().envelope(amplitude=X, duration=50)

        which for ``X`` being ``1.`` and ``0.75``, look respectively like:

        .. image:: /classes_images/rectangulars.png
            :width: 800
            :align: center
    """

    name = PulseShapeName.RECTANGULAR  # noqa: RUF045

    def envelope(self, duration: int, amplitude: float, resolution: float = 1.0) -> np.ndarray:
        """Constant amplitude envelope.

        Args:
            duration (int): Duration of the pulse (ns).
            amplitude (float): Maximum amplitude of the pulse.
            resolution (float, optional): Resolution of the pulse. Defaults to 1.

        Returns:
            ndarray: Amplitude of the envelope for each time step.
        """
        return amplitude * np.ones(round(duration / resolution))



"""Test the one qubit 2 level transpiler"""

from unittest.mock import MagicMock, patch

import pytest

import qililab as ql
from qililab.analog import Qubit2LevelTranspiler


@patch.object(ql.analog.fluqe_parameter.FluqeParameter, "foo")
@pytest.fixture(name="dummy_transpiler")
def dummy_transpiler():
    """Transpiler dummy fixture"""
    mock_delta = MagicMock(return_value=2)
    mock_delta.name = "mock_delta"
    mock_epsilon = MagicMock(return_value=3)
    mock_epsilon.name = "mock_epsilon"
    return Qubit2LevelTranspiler(epsilon_model=mock_epsilon, delta_model=mock_delta)


class TestQubit2LevelTranspiler:
    def test_init(self, dummy_transpiler):
        """Test init method"""
        assert dummy_transpiler.delta_model.name == "mock_delta"
        assert dummy_transpiler.epsilon_model.name == "mock_epsilon"
        assert dummy_transpiler.delta.name == "delta"
        assert dummy_transpiler.epsilon.name == "epsilon"
        assert dummy_transpiler.phiz.name == "phiz"
        assert dummy_transpiler.phix.name == "phix"

    def test_transpiler(self, dummy_transpiler):
        """Test transpiler"""
        # test set delta
        dummy_transpiler.delta(4)
        dummy_transpiler.delta_model.assert_called_once_with(4)
        assert dummy_transpiler.phix() == 2
        # test set epsilon
        dummy_transpiler.epsilon(5)
        dummy_transpiler.epsilon_model.assert_called_once_with(phix=2, epsilon=5)
        assert dummy_transpiler.phiz() == 3
        # test call method
        phix, phiz = dummy_transpiler(delta=1, epsilon=2)
        assert phix == 2
        assert phiz == 3



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from types import MappingProxyType
from typing import Callable

from qililab.qprogram.calibration import Calibration
from qililab.qprogram.crosstalk_matrix import CrosstalkMatrix
from qililab.qprogram.operations import ExecuteQProgram, GetParameter, SetParameter
from qililab.qprogram.operations.set_crosstalk import SetCrosstalk
from qililab.qprogram.qprogram import QProgram
from qililab.qprogram.structured_program import StructuredProgram
from qililab.qprogram.variable import Domain, Variable
from qililab.typings.enums import Parameter
from qililab.yaml import yaml


@yaml.register_class
class Experiment(StructuredProgram):
    """Represents an experiment.

    This class allows setting platform parameters and executing quantum programs.
    """

    _domain_of_parameter: MappingProxyType[Parameter, Domain] = MappingProxyType(
        {
            Parameter.AMPLITUDE: Domain.Voltage,
            Parameter.GAIN: Domain.Voltage,
            Parameter.GAIN_I: Domain.Voltage,
            Parameter.GAIN_Q: Domain.Voltage,
            Parameter.DC_OFFSET: Domain.Voltage,
            Parameter.OFFSET_I: Domain.Voltage,
            Parameter.OFFSET_Q: Domain.Voltage,
            Parameter.OFFSET_OUT0: Domain.Voltage,
            Parameter.OFFSET_OUT1: Domain.Voltage,
            Parameter.OFFSET_OUT2: Domain.Voltage,
            Parameter.OFFSET_OUT3: Domain.Voltage,
            Parameter.FLUX: Domain.Flux,
            Parameter.DURATION: Domain.Time,
            Parameter.LO_FREQUENCY: Domain.Frequency,
            Parameter.IF: Domain.Frequency,
            Parameter.PHASE: Domain.Phase,
            Parameter.DRAG_COEFFICIENT: Domain.Scalar,
            Parameter.THRESHOLD: Domain.Scalar,
            Parameter.THRESHOLD_ROTATION: Domain.Scalar,
        }
    )

    _type_of_parameter: MappingProxyType[Parameter, type] = MappingProxyType(
        {Parameter.DRAG_COEFFICIENT: float, Parameter.THRESHOLD: float, Parameter.THRESHOLD_ROTATION: float}
    )

    def __init__(self, label: str) -> None:
        super().__init__()

        self.label: str = label

    def get_parameter(self, alias: str, parameter: Parameter, channel_id: int | None = None):
        """Set a platform parameter.

        Appends a SetParameter operation to the active block of the experiment.

        Args:
            alias (str): The alias for the platform component.
            parameter (Parameter): The parameter to set.
            value (int | float): The value to set for the parameter.
        """
        variable = self.variable(
            label=f"{parameter.value} of {alias}",
            domain=self._domain_of_parameter.get(parameter, Domain.Scalar),
            type=self._type_of_parameter.get(parameter, None),
        )
        operation = GetParameter(variable=variable, alias=alias, parameter=parameter, channel_id=channel_id)
        self._active_block.append(operation)
        return variable

    def set_parameter(
        self,
        alias: str,
        parameter: Parameter,
        value: int | float | bool | Variable,
        channel_id: int | None = None,
    ):
        """Set a platform parameter.

        Appends a SetParameter operation to the active block of the experiment.

        Args:
            alias (str): The alias for the platform component.
            parameter (Parameter): The parameter to set.
            value (int | float): The value to set for the parameter.
        """
        operation = SetParameter(alias=alias, parameter=parameter, value=value, channel_id=channel_id)
        self._active_block.append(operation)

    def execute_qprogram(
        self,
        qprogram: QProgram | Callable[..., QProgram],  # type: ignore
        bus_mapping: dict[str, str] | None = None,
        debug: bool = False,
        calibration: Calibration | None = None,
    ):
        """Execute a quantum program within the experiment.

        Appends an ExecuteQProgram operation to the active block of the experiment.

        Args:
            qprogram (QProgram): The quantum program to be executed.
        """
        operation = ExecuteQProgram(qprogram=qprogram, bus_mapping=bus_mapping, calibration=calibration, debug=debug)
        self._active_block.append(operation)

    def set_crosstalk(self, crosstalk: CrosstalkMatrix):
        operation = SetCrosstalk(crosstalk=crosstalk)
        self._active_block.append(operation)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import numpy.typing as npt


def product(array: npt.NDArray | list | tuple, start: int = 0, stop: int = -1):
    """Takes in an array and does the product-of-a-sequence of the elements between `start_index` and `end_index`"""
    if stop == -1:
        stop = len(array)

    return 1 if stop - start <= 0 else np.prod(array[start:stop])


def summation(array: npt.NDArray | list | tuple, start: int = 0, stop: int = -1):
    """Takes in an array and does the summation of the elements between `start_index` and `end_index`"""
    if stop == -1:
        stop = len(array)

    return 0 if stop - start <= 0 else np.sum(array[start:stop])


def _do_target_dimensions_check(new_dimension_shape: npt.NDArray | list | tuple):
    """Checks that all target dimensions should have at least one element."""
    if any(size < 1 for size in new_dimension_shape):
        raise ValueError(
            f"Sizes of of target array should not be 0 nor negative: {new_dimension_shape!s}"
        )  # pragma: no cover


def _do_size_match_check(original_size: int, new_dimension_shape: npt.NDArray | list | tuple):
    """Checks that sizes of original and target spaces are compatible."""
    new_space_size = np.prod(new_dimension_shape)
    if new_space_size != original_size:
        raise ValueError(  # pragma: no cover
            f"Sizes of original and target arrays do not match:"
            f" original size {original_size} does not match"
            f" size of {new_dimension_shape!s}: {new_space_size}"
        )


def _coordinate_decomposition_checks(original_size: int, new_dimension_shape: npt.NDArray | list | tuple):
    """Perform some checks before applying the algorithm"""
    _do_target_dimensions_check(new_dimension_shape=new_dimension_shape)
    _do_size_match_check(original_size=original_size, new_dimension_shape=new_dimension_shape)


def _get_nth_coordinate(
    coord_elem_idx: int,
    new_indices: npt.NDArray | list | tuple,
    original_idx: int,
    new_dimension_shape: npt.NDArray | list | tuple,
    number_of_dimensions: int,
):
    """Builds precisely the i_k by using:
    i_k = ((r - sum_{p=k+1}^{D-1}(i_p*prod_{l=p+1}*{D-1}(I_l))) / (prod_{l+1}^{D-1}(I_l))) % I_k"""

    numerator_series_elements = [
        new_indices[k] * product(new_dimension_shape, start=k + 1, stop=number_of_dimensions)
        for k in range(coord_elem_idx + 1, number_of_dimensions)
    ]
    numerator = original_idx - summation(numerator_series_elements, start=coord_elem_idx + 1, stop=number_of_dimensions)
    denominator = product(new_dimension_shape, start=coord_elem_idx + 1, stop=number_of_dimensions)
    modulus = new_dimension_shape[coord_elem_idx]

    return (numerator / denominator) % modulus


def coordinate_decompose(
    new_dimension_shape: npt.NDArray | list | tuple,
    original_size: int,
    original_idx: int,
):
    """Converts an index from a 1D array to an index in an NDArray, as long as both expressions have the same sizes.
    i.e. expresses the index r in an array of R elements as the indices [n, m, l] of nested arrays with shape [N, M, L],
    as log as N*M*L==R.
    In this case, we can see that r can be expressed as r = n*M*L + m*L + l
    Mod-Dividing each side by L:
    l = r % L
    Isolating m and mod-dividing by M:
    m = ((r - l) / (M*L) ) % M
    Isolating n and mod-dividing by N:
    n = ((r - l - m*L) / (N*M*L)) % N

    In general, r can be expressed as r = sum_{k=0}^{D-1} (i_k) * prod_{l=k+1}^{D-1} (I_k), where:
     * i_k is the index in the kth dimension, for k belonging to [0, D-1]
     * I_k is the size of the kth dimension, for k belonging to [0, D-1]
     * D is the number of dimensions

    This can be inverted to get the i_k component, assuming we already know [i_{k+1}, ..., i_{D-1}]:
    i_k = ((r - sum_{p=k+1}^{D-1}(i_p*prod_{l=p+1}*{D-1}(I_l))) / (prod_{l+1}^{D-1}(I_l))) % I_k

    Args:
        new_dimension_shape: array containing the sizes of the dimensions: `{I_k}`
        original_size: number of elements in the original space, for checking compatibility
        original_idx: index of the element in the original array: `r`

    Returns: array of new indices.

    """

    _coordinate_decomposition_checks(original_size=original_size, new_dimension_shape=new_dimension_shape)

    number_of_dimensions = len(new_dimension_shape)
    new_indices = np.zeros_like(new_dimension_shape)

    for coord_elem_idx in reversed(range(len(new_dimension_shape))):
        new_indices[coord_elem_idx] = _get_nth_coordinate(
            coord_elem_idx=coord_elem_idx,
            new_indices=new_indices,
            original_idx=original_idx,
            new_dimension_shape=new_dimension_shape,
            number_of_dimensions=number_of_dimensions,
        )

    return new_indices



import pytest
from unittest.mock import MagicMock
from qililab.instruments.mini_circuits.attenuator import Attenuator
from qililab.instruments.instrument import ParameterNotFound
from qililab.typings import Parameter
from qililab.typings.instruments.mini_circuits import MiniCircuitsDriver

@pytest.fixture
def attenuator_settings():
    return {
        "alias": "attenuator_1",
        "attenuation": 10.0
    }

@pytest.fixture
def attenuator(attenuator_settings):
    attenuator = Attenuator(settings=attenuator_settings)
    attenuator.device = MagicMock(spec=MiniCircuitsDriver)
    return attenuator

class TestAttenuator:

    def test_attenuator_initialization(self, attenuator, attenuator_settings):
        assert attenuator.alias == "attenuator_1"
        assert attenuator.settings.alias == "attenuator_1"
        assert attenuator.settings.attenuation == 10.0
        assert attenuator.device is not None

    def test_attenuator_str(self, attenuator):
        assert str(attenuator) == "attenuator_1"

    def test_attenuator_attenuation_property(self, attenuator):
        assert attenuator.attenuation == 10.0

    def test_set_parameter_attenuation(self, attenuator):
        # Test setting attenuation parameter
        attenuator.set_parameter(Parameter.ATTENUATION, 15.0)
        assert attenuator.attenuation == 15.0
        attenuator.device.setup.assert_called_once_with(attenuation=15.0)

    def test_set_parameter_invalid(self, attenuator):
        # Test setting an invalid parameter
        with pytest.raises(ParameterNotFound):
            attenuator.set_parameter(Parameter.BUS_FREQUENCY, 42.0)

    def test_get_parameter_attenuation(self, attenuator):
        # Test setting attenuation parameter
        attenuator.get_parameter(Parameter.ATTENUATION)
        assert attenuator.attenuation == 10.0

    def test_get_parameter_invalid(self, attenuator):
        # Test setting an invalid parameter
        with pytest.raises(ParameterNotFound):
            attenuator.get_parameter(Parameter.BUS_FREQUENCY)

    def test_initial_setup(self, attenuator):
        # Test initial setup of the attenuator
        attenuator.initial_setup()
        attenuator.device.setup.assert_called_once_with(attenuation=attenuator.attenuation)

    def test_turn_on(self, attenuator):
        # No specific implementation in the provided code, just test invocation
        assert attenuator.turn_on() is None

    def test_turn_off(self, attenuator):
        # No specific implementation in the provided code, just test invocation
        assert attenuator.turn_off() is None

    def test_reset(self, attenuator):
        # No specific implementation in the provided code, just test invocation
        assert attenuator.reset() is None



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Qblox SPI Rack Controller class"""

from dataclasses import dataclass
from typing import Sequence

from qililab.instrument_controllers.instrument_controller import InstrumentController, InstrumentControllerSettings
from qililab.instrument_controllers.utils.instrument_controller_factory import InstrumentControllerFactory
from qililab.instruments.qblox.qblox_d5a import QbloxD5a
from qililab.instruments.qblox.qblox_s4g import QbloxS4g
from qililab.typings.enums import ConnectionName, InstrumentControllerName, InstrumentTypeName
from qililab.typings.instruments.spi_rack import SPI_Rack


@InstrumentControllerFactory.register
class QbloxSPIRackController(InstrumentController):
    """Qblox SPI Rack Controller class.

    Args:
        name (InstrumentControllerName): Name of the Instrument Controller.
        number_available_modules (int): Number of modules available in the
            Instrument Controller.
        settings (QbloxSPIRackControllerSettings): Settings of the Qblox SPI
            Rack Instrument Controller.
    """

    name = InstrumentControllerName.QBLOX_SPIRACK
    number_available_modules = 12
    device: SPI_Rack
    modules: Sequence[QbloxD5a | QbloxS4g]

    @dataclass
    class QbloxSPIRackControllerSettings(InstrumentControllerSettings):
        """Contains the settings of a specific Qblox Cluster Controller."""

        reset: bool = False

        def __post_init__(self):
            super().__post_init__()
            self.connection.name = ConnectionName.USB

    settings: QbloxSPIRackControllerSettings

    def _initialize_device(self):
        """Initialize device controller."""
        self.device = SPI_Rack(name=f"{self.name.value}_{self.alias}", address=f"/dev/{self.address}")

    def _set_device_to_all_modules(self):
        """Sets the initialized device to all attached modules,
        taking it from the Qblox Cluster device modules
        """
        for module, slot_id in zip(self.modules, self.connected_modules_slot_ids):
            self.device.add_spi_module(address=slot_id, module_type=module.name)
            module.device = self.module(module_id=slot_id)  # slot_id represents the number displayed in the cluster

    def _check_supported_modules(self):
        """check if all instrument modules loaded are supported modules for the controller."""
        for module in self.modules:
            if not isinstance(module, QbloxD5a) and not isinstance(module, QbloxS4g):
                raise ValueError(
                    f"Instrument {type(module)} not supported."
                    + f"The only supported instrument are {InstrumentTypeName.QBLOX_D5A} "
                    + f"and {InstrumentTypeName.QBLOX_S4G}."
                )

    # TODO: once the Xtalk is handled by qililab module has to be a private method
    def module(self, module_id: int):
        """get module associated to the specific module id

        Args:
            module_id (int): slot index

        Returns:
            _type_: _description_
        """
        return getattr(self.device, f"module{module_id}")



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains all the decomposition and transpilation methods used within qililab.

Transpilation
~~~~~~~~~~~~~

.. currentmodule:: qililab.digital

.. autosummary::
    :toctree: api

    ~CircuitTranspiler
    ~DigitalTranspilationConfig

"""

from .circuit_transpiler import CircuitTranspiler, DigitalTranspilationConfig
from .gate_decompositions import GateDecompositions
from .native_gates import Drag, Wait

__all__ = ["CircuitTranspiler", "DigitalTranspilationConfig", "Drag", "GateDecompositions", "Wait"]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Instruments class"""

import io
from dataclasses import dataclass

from ruamel.yaml import YAML

from qililab.instruments.instrument import Instrument


@dataclass
class Instruments:
    """Instruments class."""

    elements: list[Instrument]

    def get_instrument(self, alias: str):
        """Get element given an alias."""
        return next((element for element in self.elements if element.alias == alias), None)

    def to_dict(self):
        """Return a dict representation of the Instruments class."""
        return [instrument.to_dict() for instrument in self.elements]

    def __str__(self) -> str:
        """
        Returns:
            str: String representation of the Instruments class.
        """
        return str(YAML().dump(self._short_dict(), io.BytesIO()))

    def _short_dict(self):
        """Return a dict representation of the Instruments class discarding all static elements."""
        return [instrument.short_dict() for instrument in self.elements]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Signal processing module."""
import numpy as np
import numpy.typing as npt


def demodulate(
    i: npt.NDArray[np.float32], q: npt.NDArray[np.float32], frequency: float, phase_offset: float
) -> tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]:
    """Demodulates an (i, q) signal with a certain frequency and phase offset.

    Args:
        i (npt.NDArray[np.float32]): I component of the signal.
        q (npt.NDArray[np.float32]): Q component of the signal.
        frequency (float): demodulation frequency.
        phase_offset (float): demodulation phase offset.

    Returns:
        tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]: Demodulated signal.
    """
    time = np.arange(len(i)) * 1e-9
    cosalpha = np.cos(2 * np.pi * frequency * time + phase_offset)
    sinalpha = np.sin(2 * np.pi * frequency * time + phase_offset)
    mod_matrix = np.sqrt(2) * np.array([[cosalpha, -sinalpha], [+sinalpha, cosalpha]])
    i_demod, q_demod = np.transpose(np.einsum("abt,bt->ta", mod_matrix, [i, q]))
    return i_demod, q_demod


def modulate(
    i: npt.NDArray[np.float32], q: npt.NDArray[np.float32], frequency: float, phase_offset: float
) -> tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]:
    """Modulates an (i, q) signal with a certain frequency and phase offset.

    Args:
        i (npt.NDArray[np.float32]): I component of the signal.
        q (npt.NDArray[np.float32]): Q component of the signal.
        frequency (float): modulation frequency.
        phase_offset (float): modulation phase offset.

    Returns:
        tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]: Modulated signal.
    """
    time = np.arange(len(i)) * 1e-9
    cosalpha = np.cos(2 * np.pi * frequency * time + phase_offset)
    sinalpha = np.sin(2 * np.pi * frequency * time + phase_offset)
    mod_matrix = (1.0 / np.sqrt(2)) * np.array([[cosalpha, -sinalpha], [+sinalpha, cosalpha]])
    i_mod, q_mod = np.transpose(np.einsum("abt,bt->ta", mod_matrix, [i, q]))
    return i_mod, q_mod



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import h5py
import numpy as np


def stream_results(shape: tuple, path: str, loops: dict[str, np.ndarray]):
    """Constructs a StreamArray instance.

    This methods serves as a constructor for user interface of the StreamArray class.

    Args:
        shape (tuple): results array shape.
        path (str): path to save results.
        loops (dict[str, np.ndarray]): dictionary with each loop name in the experiment as key and numpy array as values.

    Examples:

        You have the option to save your results in real-time using this feature.
        This ensures that in the event of a runtime failure, you can still access results up to the point of failure
        from a resulting file. You have to specify the desired shape of the results, the loops to include,
        and the path to the resulting file. Then you can start saving results in real-time by doing:


        .. code-block:: python

            import numpy as np
            import qililab as ql

            LOOP_VALUES = np.linspace(0, 1, 5)
            stream_array = ql.stream_results(shape=(5, 2), loops={"loop_name": LOOP_VALUES}, path="results.h5")

            with stream_array:
                for i, value in enumerate(LOOP_VALUES):
                    stream_array[(i, 0)] = value

        >>> stream_array
        [[0.   0.  ]
         [0.25 0.  ]
         [0.5  0.  ]
         [0.75 0.  ]
         [1.   0.  ]]

        >>> ql.load_results("results.h5")
        (array([[0., 0.],
               [0.25, 0.],
               [0.5, 0.],
               [0.75, 0.],
               [1., 0.]]), {'loop_name': array([0.  , 0.25, 0.5 , 0.75, 1.  ])})

    .. note::

        The output of `stream_results` is not picklable, thus it cannot be returned in a SLURM job. In this scenario we suggest copying the array with the data into a different variable and returning that variable instead:

        .. code-block:: python

            %%submit_job -o results -d galadriel
            LOOP_VALUES = np.linspace(0, 1, 5)
            stream_array = ql.stream_results(shape=(5, 2), loops={"loop_name": LOOP_VALUES}, path="results.h5")
            with stream_array:
                for i, value in enumerate(LOOP_VALUES):
                    # Here you can execute any algorithm you want
                    stream_array[(i, 0)] = value

            results = stream_array.results

        >>> results.result()
        (array([[0.   0.  ]
                [0.25 0.  ]
                [0.5  0.  ]
                [0.75 0.  ]
                [1.   0.  ]])

    """
    return StreamArray(shape=shape, path=path, loops=loops)


class StreamArray:
    """
    Allows for real time saving of results from an experiment.

    This class wraps a numpy array and adds a context manager to save results on real time while they are acquired by
    the instruments.

    Args:
        shape (tuple): results array shape.
        path (str): path to save results.
        loops (dict[str, np.ndarray]): dictionary with each loop name in the experiment as key and numpy array as values.
    """

    def __init__(self, shape: tuple, path: str, loops: dict[str, np.ndarray]):
        self.results = np.zeros(shape=shape)
        self.path = path
        self.loops = loops
        self._file: h5py.File | None = None
        self._dataset = None

    def __setitem__(self, key: tuple, value: float):
        """Sets and item by key and value in the dataset.

        Args:
            key (tuple): key for the item to save.
            value (float): value to save.
        """
        if self._file is not None and self._dataset is not None:
            self._dataset[key] = value
        self.results[key] = value

    def __enter__(self):
        self._file = h5py.File(name=self.path, mode="w")
        # Save loops
        g = self._file.create_group(name="loops")
        for loop_name, array in self.loops.items():
            g.create_dataset(name=loop_name, data=array)
        # Save results
        self._dataset = self._file.create_dataset("results", data=self.results)

        return self

    def __exit__(self, *args):
        """Exits the context manager."""
        if self._file is not None:
            self._file.__exit__()
            self._file = None

    def __getitem__(self, index: int):
        """Gets item by index.

        Args:
            index (int): item's index.
        """
        return self.results[index]

    def __len__(self):
        """Gets length of results."""
        return len(self.results)

    def __iter__(self):
        """Gets iterator of results."""
        return iter(self.results)

    def __str__(self):
        """Gets string representation of results."""
        return str(self.results)

    def __repr__(self):
        """Gets string representation of results."""
        return repr(self.results)

    def __contains__(self, item: dict[str, Any]):
        """Returns if an item is contained in results.

        Args:
            item (dict): item

        Returns:
            bool: True if an item is contained in results.
        """
        return item in self.results



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Keithley2600 Instrument Controller """
from dataclasses import dataclass
from typing import Sequence

from qililab.instrument_controllers.instrument_controller import InstrumentControllerSettings
from qililab.instrument_controllers.single_instrument_controller import SingleInstrumentController
from qililab.instrument_controllers.utils.instrument_controller_factory import InstrumentControllerFactory
from qililab.instruments.keithley.keithley_2600 import Keithley2600
from qililab.typings import Keithley2600Driver
from qililab.typings.enums import ConnectionName, InstrumentControllerName, InstrumentTypeName


@InstrumentControllerFactory.register
class Keithley2600Controller(SingleInstrumentController):
    """Keithley2600 class

    Args:
        name (InstrumentControllerName): Name of the Instrument Controller.
        device (Keithley2600Driver): Instance of the qcodes Keithley2600 class.
        settings (Keithley2600Settings): Settings of the instrument.
    """

    name = InstrumentControllerName.KEITHLEY2600
    device: Keithley2600Driver
    modules: Sequence[Keithley2600]

    @dataclass
    class Keithley2600ControllerSettings(InstrumentControllerSettings):
        """Contains the settings of a specific Keithley2600 Controller."""

        def __post_init__(self):
            super().__post_init__()
            self.connection.name = ConnectionName.TCP_IP

    settings: Keithley2600ControllerSettings

    def _initialize_device(self):
        """Initialize device attribute to the corresponding device class."""
        self.device = Keithley2600Driver(
            name=f"{self.name.value}_{self.alias}", address=f"TCPIP0::{self.address}::INSTR", visalib="@py"
        )

    def _check_supported_modules(self):
        """check if all instrument modules loaded are supported modules for the controller."""
        for module in self.modules:
            if not isinstance(module, Keithley2600):
                raise ValueError(
                    f"Instrument {type(module)} not supported."
                    + f"The only supported instrument is {InstrumentTypeName.KEITHLEY2600}"
                )



import pytest
from unittest.mock import MagicMock
from qililab.instruments import Instrument
from qililab.typings import Parameter, ParameterValue, ChannelID

# A concrete subclass of Instrument for testing purposes
class DummyInstrument(Instrument):
    def turn_on(self):
        return "Instrument turned on"

    def turn_off(self):
        return "Instrument turned off"

    def reset(self):
        return "Instrument reset"

    def initial_setup(self):
        return "Initial Setup"

    def get_parameter(self, parameter: Parameter, channel_id: ChannelID | None = None) -> ParameterValue:
        return "parameter_value"

    def set_parameter(self, parameter: Parameter, value: ParameterValue, channel_id: ChannelID | None = None):
        return True

@pytest.fixture
def instrument_settings():
    return {
        "alias": "test_instrument",
    }

@pytest.fixture
def instrument(instrument_settings):
    return DummyInstrument(settings=instrument_settings)

class TestInstrumentBase:
    """Test class for the Instrument abstract class, ensuring common functionality is tested."""

    def test_instrument_initialization(self, instrument, instrument_settings):
        assert instrument.alias == "test_instrument"
        assert instrument.settings.alias == "test_instrument"

    def test_instrument_str(self, instrument):
        assert str(instrument) == "test_instrument"

    def test_instrument_is_device_active(self, instrument):
        # Device is initially not set, so should return False
        assert instrument.is_device_active() is False

        # After setting a device, it should return True
        instrument.device = MagicMock()
        assert instrument.is_device_active() is True

    def test_instrument_turn_on(self, instrument):
        assert instrument.turn_on() == "Instrument turned on"

    def test_instrument_turn_off(self, instrument):
        assert instrument.turn_off() == "Instrument turned off"

    def test_instrument_reset(self, instrument):
        assert instrument.reset() == "Instrument reset"

    def test_instrument_get_parameter(self, instrument):
        parameter = MagicMock(spec=Parameter)
        assert instrument.get_parameter(parameter) == "parameter_value"

    def test_instrument_set_parameter(self, instrument):
        parameter = MagicMock(spec=Parameter)
        value = 42
        assert instrument.set_parameter(parameter, value) is True

    def test_instrument_is_awg(self, instrument):
        assert instrument.is_awg() is False  # Default implementation returns False

    def test_instrument_is_adc(self, instrument):
        assert instrument.is_adc() is False  # Default implementation returns False



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""VoltageSource class."""

from dataclasses import dataclass

from qililab.instruments.instrument import Instrument


class VoltageSource(Instrument):
    """Abstract base class defining all instruments used to generate voltages."""

    @dataclass
    class VoltageSourceSettings(Instrument.InstrumentSettings):
        """Contains the settings of a specific signal generator.

        Args:
            voltage (list[float]): List of voltages of the instrument's DACs in V. Value range is (-40, 40).
            span (list[str]): A list of the max range of each DACs.
            ramping_enabled (list[bool]): A list of ramping enabled boolean values per DAC.
            ramp_rate (list[float]): List of ramping rates in mA / ms per DAC.
            dacs (list[int]): List of the dacs of which to set the current.
        """

        voltage: list[float]
        span: list[str]
        ramping_enabled: list[bool]
        ramp_rate: list[float]
        dacs: list[int]  # indices of the dacs to use

    settings: VoltageSourceSettings

    @property
    def voltage(self):
        """VoltageSource 'voltage' property.

        Returns:
            float: settings.voltage.
        """
        return self.settings.voltage

    @property
    def dacs(self):
        """CurrentSource 'dacs' property.

        Returns:
            int: settings.dacs
        """
        return self.settings.dacs

    @property
    def span(self):
        """VoltageSource 'span' property.

        Returns:
            float: settings.span.
        """
        return self.settings.span

    @property
    def ramping_enabled(self):
        """VoltageSource 'ramping_enabled' property.

        Returns:
            float: settings.ramping_enabled.
        """
        return self.settings.ramping_enabled

    @property
    def ramp_rate(self):
        """VoltageSource 'ramp_rate' property.

        Returns:
            float: settings.ramp_rate.
        """
        return self.settings.ramp_rate

    def to_dict(self):
        """Return a dict representation of the VoltageSource class."""
        return dict(super().to_dict().items())



import copy
from unittest import mock
from unittest.mock import MagicMock, patch

import pytest

from qililab.instrument_controllers.qblox.qblox_spi_rack_controller import QbloxSPIRackController
from qililab.instruments.qblox import QbloxS4g
from qililab.platform import Platform
from tests.data import SauronSpiRack
from tests.test_utils import build_platform


@pytest.fixture(name="platform")
def fixture_platform() -> Platform:
    """Return Platform object."""
    return build_platform(runcard=SauronSpiRack.runcard)


class TestQbloxSpiRackController:
    """Unit tests checking the SPI rack controller attributes and methods"""

    def test_initialization(self, platform: Platform):
        """Test SPI rack controller has been initialized correctly."""
        controller_instance = platform.instrument_controllers.get_instrument_controller(alias="spi_controller_usb")
        assert isinstance(controller_instance, QbloxSPIRackController)

        controller_settings = controller_instance.settings
        assert isinstance(controller_settings, QbloxSPIRackController.QbloxSPIRackControllerSettings)

        controller_modules = controller_instance.modules
        assert len(controller_modules) == 1
        assert isinstance(controller_modules[0], QbloxS4g)

    @patch("qililab.instrument_controllers.qblox.qblox_spi_rack_controller.SPI_Rack", autospec=True)
    def test_initialize_device(self, device_mock: MagicMock, platform: Platform):
        """Test SPI rack controller initializes device correctly."""
        controller_instance = platform.instrument_controllers.get_instrument_controller(alias="spi_controller_usb")

        controller_instance._initialize_device()

        device_mock.assert_called_once_with(
            name=f"{controller_instance.name.value}_{controller_instance.alias}",
            address=f"/dev/{controller_instance.address}",
        )

    @patch("qililab.instrument_controllers.qblox.qblox_spi_rack_controller.SPI_Rack", autospec=True)
    def test_set_device_to_all_modules(self, device_mock: MagicMock, platform: Platform):
        """Test SPI rack controller initializes device correctly foa all modules."""
        controller_instance = platform.instrument_controllers.get_instrument_controller(alias="spi_controller_usb")

        controller_instance._initialize_device()
        controller_instance.device = mock.Mock()
        controller_instance._set_device_to_all_modules()

        controller_instance.device.add_spi_module.assert_called_once()

    @patch("qililab.instrument_controllers.qblox.qblox_spi_rack_controller.SPI_Rack", autospec=True)
    def test_module(self, device_mock: MagicMock, platform: Platform):
        """Test SPI rack controller sets module correctly."""
        controller_instance = platform.instrument_controllers.get_instrument_controller(alias="spi_controller_usb")

        controller_instance._initialize_device()
        controller_instance.device = mock.Mock()
        controller_instance.module(1)

        assert controller_instance.module(1) == getattr(controller_instance.device, f"module{1}")

    def test_check_supported_modules_raises_exception(self, platform: Platform):
        """Test SPI rack controller raises an error if initialized with wrong module."""
        controller_instance = platform.instrument_controllers.get_instrument_controller(
            alias="spi_controller_wrong_module"
        )
        with pytest.raises(ValueError):
            controller_instance._check_supported_modules()



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from qililab.qprogram.element import Element
from qililab.qprogram.variable import Variable


class Operation(Element):
    def get_variables(self) -> set[Variable]:
        """Get a set of the variables used in operation, if any.

        Returns:
            set[Variable]: The set of variables used in operation.
        """
        return {attribute for attribute in self.__dict__.values() if isinstance(attribute, Variable)}
